import typing as t
from collections import namedtuple

import marshmallow
import pydantic
import pytest
import typesystem
import typesystem.fields


@pytest.fixture(scope="function")
def foo_schema(app):
    if app.schema.schema_library.lib == pydantic:
        schema = pydantic.create_model("Foo", name=(str, ...), __module__="pydantic.main")
        name = "pydantic.main.Foo"
    elif app.schema.schema_library.lib == typesystem:
        schema = typesystem.Schema(title="Foo", fields={"name": typesystem.fields.String()})
        name = "typesystem.schemas.Foo"
    elif app.schema.schema_library.lib == marshmallow:
        schema = type("Foo", (marshmallow.Schema,), {"name": marshmallow.fields.String()})
        name = "abc.Foo"
    else:
        raise ValueError(f"Wrong schema lib: {app.schema.schema_library.lib}")
    return namedtuple("FooSchema", ("schema", "name"))(schema=schema, name=name)


@pytest.fixture(scope="function")
def bar_schema(app, foo_schema):
    child_schema = foo_schema.schema
    if app.schema.schema_library.lib == pydantic:
        schema = pydantic.create_model("Bar", foo=(child_schema, ...), __module__="pydantic.main")
        name = "pydantic.main.Bar"
    elif app.schema.schema_library.lib == typesystem:
        schema = typesystem.Schema(
            title="Bar",
            fields={"foo": typesystem.Reference(to="Foo", definitions=typesystem.Definitions({"Foo": child_schema}))},
        )
        name = "typesystem.schemas.Bar"
    elif app.schema.schema_library.lib == marshmallow:
        schema = type("Bar", (marshmallow.Schema,), {"foo": marshmallow.fields.Nested(child_schema())})
        name = "abc.Bar"
    else:
        raise ValueError(f"Wrong schema lib: {app.schema.schema_library.lib}")
    return namedtuple("BarSchema", ("schema", "name"))(schema=schema, name=name)


@pytest.fixture(scope="function")
def bar_optional_schema(app, foo_schema):
    child_schema = foo_schema.schema
    if app.schema.schema_library.lib == pydantic:
        schema = pydantic.create_model(
            "BarOptional", foo=(t.Union[child_schema, None], None), __module__="pydantic.main"
        )
        name = "pydantic.main.BarOptional"
    elif app.schema.schema_library.lib == typesystem:
        schema = typesystem.Schema(
            title="BarOptional",
            fields={
                "foo": typesystem.Reference(
                    to="Foo", definitions=typesystem.Definitions({"Foo": child_schema}), allow_null=True, default=None
                )
            },
        )
        name = "typesystem.schemas.BarOptional"
    elif app.schema.schema_library.lib == marshmallow:
        schema = type(
            "BarOptional",
            (marshmallow.Schema,),
            {"foo": marshmallow.fields.Nested(child_schema(), required=False, dump_default=None, allow_none=True)},
        )
        name = "abc.BarOptional"
    else:
        raise ValueError(f"Wrong schema lib: {app.schema.schema_library.lib}")
    return namedtuple("BarOptionalSchema", ("schema", "name"))(schema=schema, name=name)


@pytest.fixture(scope="function")
def bar_list_schema(app, foo_schema):
    child_schema = foo_schema.schema
    if app.schema.schema_library.lib == pydantic:
        schema = pydantic.create_model("BarList", foo=(list[child_schema], ...), __module__="pydantic.main")
        name = "pydantic.main.BarList"
    elif app.schema.schema_library.lib == typesystem:
        schema = typesystem.Schema(
            title="BarList",
            fields={
                "foo": typesystem.Array(
                    typesystem.Reference(to="Foo", definitions=typesystem.Definitions({"Foo": child_schema}))
                )
            },
        )
        name = "typesystem.schemas.BarList"
    elif app.schema.schema_library.lib == marshmallow:
        schema = type(
            "BarList",
            (marshmallow.Schema,),
            {"foo": marshmallow.fields.List(marshmallow.fields.Nested(child_schema()))},
        )
        name = "abc.BarList"
    else:
        raise ValueError(f"Wrong schema lib: {app.schema.schema_library.lib}")
    return namedtuple("BarListSchema", ("schema", "name"))(schema=schema, name=name)


@pytest.fixture(scope="function")
def bar_dict_schema(app, foo_schema):
    child_schema = foo_schema.schema
    if app.schema.schema_library.lib == pydantic:
        schema = pydantic.create_model("BarDict", foo=(dict[str, child_schema], ...), __module__="pydantic.main")
        name = "pydantic.main.BarDict"
    elif app.schema.schema_library.lib == typesystem:
        schema = typesystem.Schema(
            title="BarDict",
            fields={
                "foo": typesystem.Object(
                    properties=typesystem.Reference(to="Foo", definitions=typesystem.Definitions({"Foo": child_schema}))
                )
            },
        )
        name = "typesystem.schemas.BarDict"
    elif app.schema.schema_library.lib == marshmallow:
        schema = type(
            "BarDict",
            (marshmallow.Schema,),
            {"foo": marshmallow.fields.Dict(values=marshmallow.fields.Nested(child_schema()))},
        )
        name = "abc.BarDict"
    else:
        raise ValueError(f"Wrong schema lib: {app.schema.schema_library.lib}")
    return namedtuple("BarDictSchema", ("schema", "name"))(schema=schema, name=name)


@pytest.fixture(scope="function")
def schemas(foo_schema, bar_schema, bar_list_schema, bar_dict_schema):
    return {"Foo": foo_schema, "Bar": bar_schema, "BarList": bar_list_schema, "BarDict": bar_dict_schema}



import typing as t

from flama.ddd.workers.base import BaseWorker

if t.TYPE_CHECKING:
    from flama import Flama
    from flama.client import Client

__all__ = ["HTTPWorker"]


class HTTPWorker(BaseWorker):
    """Worker for HTTP client.

    It will provide a flama Client and create the repositories for the corresponding resources.
    """

    _client: "Client"

    def __init__(self, url: t.Union[str, t.Callable[[], str]], app: t.Optional["Flama"] = None, **client_kwargs: t.Any):
        super().__init__(app=app)
        self._url = url
        self._client_kwargs = client_kwargs

    @property
    def url(self) -> str:
        return self._url() if callable(self._url) else self._url

    @property
    def client(self) -> "Client":
        """Client to interact with an HTTP resource

        :return: Flama client.
        :raises AttributeError: If the client is not initialized.
        """
        try:
            return self._client
        except AttributeError:
            raise AttributeError("Client not initialized")

    @client.setter
    def client(self, client: "Client") -> None:
        """Set the client to interact with an HTTP resource.

        :param client: Flama client.
        """
        self._client = client

    @client.deleter
    def client(self) -> None:
        """Delete the client."""
        del self._client

    async def set_up(self) -> None:
        """Initialize the client with the URL."""
        from flama.client import Client

        self.client = Client(base_url=self.url, **self._client_kwargs)

        await self.client.__aenter__()

    async def tear_down(self, *, rollback: bool = False) -> None:
        """Close and delete the client.

        :param rollback: If the unit of work should be rolled back.
        """
        await self.client.__aexit__()
        del self.client

    async def repository_params(self) -> tuple[list[t.Any], dict[str, t.Any]]:
        """Get the parameters for initialising the repositories.

        :return: Parameters for initialising the repositories.
        """
        return [self.client], {}

    async def commit(self) -> None: ...

    async def rollback(self) -> None: ...



import pytest
import sqlalchemy

from flama.applications import Flama
from flama.ddd.repositories.sqlalchemy import SQLAlchemyRepository
from flama.resources import data_structures
from flama.resources.crud import CRUDResource
from flama.resources.routing import resource_method
from flama.sqlalchemy import SQLAlchemyModule, metadata


@pytest.fixture
def app(app):
    return Flama(schema=None, docs=None, modules={SQLAlchemyModule("sqlite+aiosqlite://")})


class TestCaseBaseResource:
    @pytest.fixture(scope="function")
    def resource(self, puppy_model, puppy_schema):
        class PuppyResource(CRUDResource):
            name = "puppy"
            verbose_name = "Puppy"

            model = puppy_model
            schema = puppy_schema

        return PuppyResource

    def test_meta_attributes(self, resource, puppy_model, puppy_schema):
        assert not hasattr(resource, "name")
        assert not hasattr(resource, "verbose_name")
        assert not hasattr(resource, "schema")
        assert hasattr(resource, "model")
        assert isinstance(getattr(resource, "model"), sqlalchemy.Table)
        assert hasattr(resource, "_meta")
        assert resource._meta.name == "puppy"
        assert resource._meta.verbose_name == "Puppy"

        namespaces = resource._meta.namespaces
        ddd_namespace = namespaces.pop("ddd")

        assert list(ddd_namespace.keys()) == ["repository"]
        assert issubclass(ddd_namespace["repository"], SQLAlchemyRepository)
        assert ddd_namespace["repository"].__name__ == "PuppyResourceRepository"

        assert namespaces == {
            "rest": {
                "model": data_structures.Model(
                    table=puppy_model, primary_key=data_structures.PrimaryKey(name="custom_id", type=int)
                ),
                "schemas": data_structures.Schemas(
                    input=data_structures.Schema(name="PuppyResource", schema=puppy_schema),
                    output=data_structures.Schema(name="PuppyResource", schema=puppy_schema),
                ),
            }
        }

    def test_override_method(self, app, resource):
        class SpecializedPuppyResource(resource):
            @resource_method("/")
            def list(self):
                return ["foo", "bar"]

        assert hasattr(SpecializedPuppyResource, "create")
        assert hasattr(SpecializedPuppyResource, "retrieve")
        assert hasattr(SpecializedPuppyResource, "update")
        assert hasattr(SpecializedPuppyResource, "partial_update")
        assert hasattr(SpecializedPuppyResource, "delete")
        assert hasattr(SpecializedPuppyResource, "list")
        assert hasattr(SpecializedPuppyResource, "replace")
        assert hasattr(SpecializedPuppyResource, "partial_replace")
        assert hasattr(SpecializedPuppyResource, "drop")
        assert len(SpecializedPuppyResource.routes) == 9

        assert SpecializedPuppyResource().list() == ["foo", "bar"]

    def test_new_no_model(self, puppy_schema):
        with pytest.raises(AttributeError, match=r"PuppyResource needs to define attribute 'model'"):

            class PuppyResource(CRUDResource):
                schema = puppy_schema

    def test_invalid_no_model(self, puppy_schema):
        with pytest.raises(
            AttributeError, match=r"PuppyResource model must be a valid SQLAlchemy Table instance or a Model instance"
        ):

            class PuppyResource(CRUDResource):
                schema = puppy_schema
                model = None

    def test_new_no_name(self, puppy_model, puppy_schema):
        class PuppyResource(CRUDResource):
            model = puppy_model
            schema = puppy_schema

        assert PuppyResource._meta.name == "PuppyResource"

    def test_new_wrong_name(self, puppy_model, puppy_schema):
        with pytest.raises(AttributeError, match=r"PuppyResource invalid resource name '123foo'"):

            class PuppyResource(CRUDResource):
                model = puppy_model
                schema = puppy_schema
                name = "123foo"

    def test_new_no_schema(self, puppy_model):
        with pytest.raises(
            AttributeError,
            match=r"PuppyResource needs to define attribute 'schema' or the pair 'input_schema' and 'output_schema'",
        ):

            class PuppyResource(CRUDResource):
                model = puppy_model

    def test_new_no_input_schema(self, puppy_model, puppy_schema):
        with pytest.raises(
            AttributeError,
            match=r"PuppyResource needs to define attribute 'schema' or the pair 'input_schema' and 'output_schema'",
        ):

            class PuppyResource(CRUDResource):
                model = puppy_model
                output_schema = puppy_schema

    def test_new_no_output_schema(self, puppy_model, puppy_schema):
        with pytest.raises(
            AttributeError,
            match=r"PuppyResource needs to define attribute 'schema' or the pair 'input_schema' and 'output_schema'",
        ):

            class PuppyResource(CRUDResource):
                model = puppy_model
                input_schema = puppy_schema

    def test_resource_model_no_pk(self, puppy_schema):
        model_ = sqlalchemy.Table("no_pk", metadata, sqlalchemy.Column("integer", sqlalchemy.Integer))

        with pytest.raises(AttributeError, match=r"PuppyResource model must define a single-column primary key"):

            class PuppyResource(CRUDResource):
                model = model_
                schema = puppy_schema

    def test_resource_model_multicolumn_pk(self, puppy_schema):
        model_ = sqlalchemy.Table(
            "multicolumn_pk",
            metadata,
            sqlalchemy.Column("integer", sqlalchemy.Integer),
            sqlalchemy.Column("string", sqlalchemy.String),
            sqlalchemy.PrimaryKeyConstraint("integer", "string"),
        )

        with pytest.raises(AttributeError, match=r"PuppyResource model must define a single-column primary key"):

            class PuppyResource(CRUDResource):
                model = model_
                schema = puppy_schema

    def test_resource_model_invalid_type_pk(self, puppy_schema):
        model_ = sqlalchemy.Table(
            "invalid_pk", metadata, sqlalchemy.Column("id", sqlalchemy.PickleType, primary_key=True)
        )

        with pytest.raises(AttributeError, match=r"PuppyResource model primary key wrong type"):

            class PuppyResource(CRUDResource):
                model = model_
                schema = puppy_schema


class TestCaseResourceMethod:
    def test_resource_method(self):
        @resource_method(path="/", methods=["POST"], name="foo", tags={"additional": "bar"})
        def foo(x: int):
            return x

        assert hasattr(foo, "_meta")
        assert foo._meta.path == "/"
        assert foo._meta.methods == {"POST"}
        assert foo._meta.name == "foo"
        assert foo._meta.tags == {"additional": "bar"}



from unittest.mock import AsyncMock, MagicMock, PropertyMock, call, patch

import pytest
import starlette.websockets

from flama import Component, Flama, endpoints, exceptions, types, websockets


class Puppy:
    name = "Canna"


class PuppyComponent(Component):
    def resolve(self) -> Puppy:
        return Puppy()


class TestCaseWebSocketEndpoint:
    @pytest.fixture(scope="class")
    def app(self, app):
        return Flama(schema=None, docs=None, components=[PuppyComponent()])

    @pytest.fixture
    def endpoint(self, app, asgi_scope, asgi_receive, asgi_send):
        @app.websocket_route("/")
        class FooEndpoint(endpoints.WebSocketEndpoint):
            def get(self): ...

        asgi_scope["app"] = app
        asgi_scope["root_app"] = app
        asgi_scope["type"] = "websocket"
        with patch("flama.websockets.WebSocket", spec=websockets.WebSocket):
            return FooEndpoint(asgi_scope, asgi_receive, asgi_send)

    @pytest.mark.skip(reason="Cannot test websockets with current client")  # CAVEAT: Client doesn't support websockets
    @pytest.mark.parametrize(
        ["encoding", "send_method", "data", "expected_result"],
        (
            pytest.param("bytes", "send_bytes", b"foo", {"bytes": b"foo", "type": "websocket.send"}, id="bytes"),
            pytest.param(
                "bytes", "send_text", "foo", {"code": 1003, "type": "websocket.close", "reason": ""}, id="bytes_wrong"
            ),
            pytest.param("text", "send_text", "foo", {"text": "foo", "type": "websocket.send"}, id="text"),
            pytest.param(
                "text", "send_bytes", b"foo", {"code": 1003, "type": "websocket.close", "reason": ""}, id="text_wrong"
            ),
            pytest.param(
                "json", "send_json", {"foo": "bar"}, {"text": '{"foo":"bar"}', "type": "websocket.send"}, id="json"
            ),
            pytest.param(
                "json",
                "send_bytes",
                b'{"foo": "bar"}',
                {"text": '{"foo":"bar"}', "type": "websocket.send"},
                id="json_using_bytes",
            ),
            pytest.param(
                "json",
                "send_text",
                b'{"foo":"bar"}',
                {"text": '{"foo":"bar"}', "type": "websocket.send"},
                id="json_using_text",
            ),
            pytest.param(
                "json", "send_bytes", b"foo", {"code": 1003, "type": "websocket.close", "reason": ""}, id="json_wrong"
            ),
            pytest.param(
                None,
                "send_bytes",
                b"foo",
                {"bytes": b"foo", "type": "websocket.send"},
                id="default_encoding",
            ),
            pytest.param(
                "unknown",
                "send_bytes",
                b"foo",
                {"code": 1003, "type": "websocket.close", "reason": ""},
                id="unknown_encoding",
            ),
        ),
    )
    def test_receive(self, app, client, encoding, send_method, data, expected_result):
        encoding_ = encoding

        @app.websocket_route("/")
        class FooWebSocketEndpoint(endpoints.WebSocketEndpoint):
            encoding = encoding_

            async def on_receive(self, websocket: websockets.WebSocket, data: types.Data):
                await getattr(websocket, f"send_{encoding_ or 'bytes'}")(data)

        with client.websocket_connect("/") as ws:
            getattr(ws, send_method)(data)
            result = ws.receive()

        assert result == expected_result

    @pytest.mark.skip(reason="Cannot test websockets with current client")  # CAVEAT: Client doesn't support websockets
    def test_injecting_component(self, app, client):
        @app.websocket_route("/")
        class FooWebSocketEndpoint(endpoints.WebSocketEndpoint):
            encoding = types.Encoding("bytes")

            async def on_connect(self, websocket: websockets.WebSocket):
                await websocket.accept()

            async def on_receive(self, websocket: websockets.WebSocket, data: types.Data, puppy: Puppy):
                await websocket.send_json({"puppy": puppy.name})

            async def on_disconnect(self, websocket: websockets.WebSocket, websocket_code: types.Code):
                pass

        with client.websocket_connect("/") as ws:
            ws.send_bytes(b"")
            result = ws.receive_json()

        assert result == {"puppy": "Canna"}

    @pytest.mark.skip(reason="Cannot test websockets with current client")  # CAVEAT: Client doesn't support websockets
    def test_fail_connecting(self, app, client):
        @app.websocket_route("/")
        class FooWebSocketEndpoint(endpoints.WebSocketEndpoint):
            async def on_connect(self, websocket: websockets.WebSocket):
                raise Exception("Error connecting socket")

        with pytest.raises(Exception, match="Error connecting socket"), client.websocket_connect("/") as ws:
            ws.send_bytes("foo")

    @pytest.mark.skip(reason="Cannot test websockets with current client")  # CAVEAT: Client doesn't support websockets
    def test_fail_receiving(self, app, client):
        @app.websocket_route("/")
        class FooWebSocketEndpoint(endpoints.WebSocketEndpoint):
            async def on_receive(self, websocket: websockets.WebSocket, data: types.Data):
                raise ValueError("Foo")

        with pytest.raises(ValueError, match="Foo"), client.websocket_connect("/") as ws:
            ws.send_bytes("foo")
            result = ws.receive()

            assert result == {"code": 1011, "type": "websocket.close", "reason": ""}

    def test_init(self, app, asgi_scope, asgi_receive, asgi_send):
        with patch("flama.websockets.WebSocket") as websocket_mock:
            route = app.add_websocket_route("/", endpoints.WebSocketEndpoint)
            asgi_scope = types.Scope(
                {
                    **asgi_scope,
                    "app": app,
                    "root_app": app,
                    "type": "websocket",
                    "path": "/",
                    "path_params": {},
                    "endpoint": endpoints.WebSocketEndpoint,
                    "route": route,
                }
            )
            endpoint = endpoints.WebSocketEndpoint(asgi_scope, asgi_receive, asgi_send)
            assert endpoint.state == {
                "scope": asgi_scope,
                "receive": asgi_receive,
                "send": asgi_send,
                "exc": None,
                "app": app,
                "root_app": app,
                "path_params": {},
                "route": route,
                "websocket": websocket_mock(),
                "websocket_code": None,
                "websocket_encoding": None,
                "websocket_message": None,
            }

    def test_allowed_handlers(self, endpoint):
        assert endpoint.allowed_handlers() == {
            "WEBSOCKET_CONNECT": endpoint.__class__.on_connect,
            "WEBSOCKET_RECEIVE": endpoint.__class__.on_receive,
            "WEBSOCKET_DISCONNECT": endpoint.__class__.on_disconnect,
        }

    @pytest.mark.parametrize(
        ["endpoint_receive", "websocket_receive", "exception", "result_code", "result_message"],
        (
            pytest.param(
                [None, None, None],
                [
                    {"type": "websocket.receive", "code": 1000, "bytes": "foo"},
                    {"type": "websocket.disconnect", "code": 1000},
                ],
                None,
                1000,
                {"type": "websocket.disconnect", "code": 1000},
                id="ok",
            ),
            pytest.param(
                [None, None, None],
                [
                    {"type": "websocket.receive", "code": 1000, "bytes": "foo"},
                    starlette.websockets.WebSocketDisconnect(1006, "Abnormal Closure"),
                ],
                exceptions.WebSocketException(1006, "Abnormal Closure"),
                1006,
                {"type": "websocket.receive", "code": 1000, "bytes": "foo"},
                id="disconnect",
            ),
            pytest.param(
                [None, exceptions.WebSocketException(1003, "Unsupported Data"), None],
                [{"type": "websocket.receive", "code": 1000, "bytes": "foo"}],
                exceptions.WebSocketException(1003, "Unsupported Data"),
                1003,
                {"type": "websocket.receive", "code": 1000, "bytes": "foo"},
                id="websocket_exception",
            ),
            pytest.param(
                [None, ValueError("Foo"), None],
                [{"type": "websocket.receive", "code": 1000, "bytes": "foo"}],
                ValueError("Foo"),
                1011,
                {"type": "websocket.receive", "code": 1000, "bytes": "foo"},
                id="exception",
            ),
        ),
        indirect=["exception"],
    )
    async def test_dispatch(
        self, app, endpoint, endpoint_receive, websocket_receive, exception, result_code, result_message
    ):
        app.injector.inject = AsyncMock(side_effect=[AsyncMock(side_effect=x) for x in endpoint_receive])
        endpoint.state["websocket"].receive = AsyncMock(side_effect=websocket_receive)
        type(endpoint.state["websocket"]).is_connected = PropertyMock(side_effect=[True, False])

        with exception:
            await endpoint.dispatch()

        assert endpoint.state["websocket_code"] == result_code
        assert endpoint.state["websocket_message"] == result_message

    async def test_on_connect(self, endpoint):
        websocket = MagicMock(websockets.WebSocket)

        await endpoint.on_connect(websocket)

        assert websocket.accept.call_args_list == [call()]

    async def test_on_receive(self, endpoint):
        websocket = MagicMock(websockets.WebSocket)

        await endpoint.on_receive(websocket, b"foo")

    async def test_on_disconnect(self, endpoint):
        websocket = MagicMock(websockets.WebSocket)

        await endpoint.on_disconnect(websocket, types.Code(1000))

        assert websocket.close.call_args_list == [call(types.Code(1000))]



import flama

app = flama.Flama(
    openapi={
        "info": {
            "title": "Hello-🔥",
            "version": "1.0",
            "description": "My first API",
        },
        "tags": [
            {"name": "Salute", "description": "This is the salute description"},
        ],
    }
)


@app.route("/")
def home():
    """
    tags:
        - Salute
    summary:
        Returns a warming message.
    description:
        This is a more detailed description of the method itself.
        Here we can give all the details required and they will appear
        automatically in the auto-generated docs.
    responses:
        200:
            description: Warming hello message!
    """
    return {"message": "Hello 🔥"}


if __name__ == "__main__":
    flama.run(flama_app=app, server_host="0.0.0.0", server_port=8080)



import http
import uuid
from unittest.mock import MagicMock, Mock, call, patch

import httpx
import pytest

from flama import Flama
from flama.client import Client
from flama.ddd import exceptions
from flama.ddd.repositories import HTTPRepository, HTTPResourceManager, HTTPResourceRepository
from flama.sqlalchemy import SQLAlchemyModule


@pytest.fixture(scope="function")
def app():
    return Flama(schema=None, docs=None, modules={SQLAlchemyModule("sqlite+aiosqlite://")})


class TestCaseHTTPRepository:
    @pytest.fixture(scope="function")
    def client(self):
        return Mock(spec=Client)

    async def test_init(self, client):
        class Repository(HTTPRepository): ...

        repository = Repository(client)

        assert repository._client == client

    def test_eq(self, client):
        assert HTTPRepository(client) == HTTPRepository(client)
        assert HTTPRepository(client) != HTTPRepository(Mock())


class TestCaseHTTPResourceManager:
    @pytest.fixture(scope="function")
    def client(self):
        return Mock(spec=Client)

    @pytest.fixture(scope="function")
    def resource_manager(self, client):
        return HTTPResourceManager("foo", client)

    async def test_init(self, client):
        resource_manager = HTTPResourceManager("foo", client)

        assert resource_manager._client == client
        assert resource_manager.resource == "foo"

    async def test_eq(self, resource_manager, client):
        assert resource_manager == HTTPResourceManager("foo", client)
        assert resource_manager != HTTPResourceManager("foo", Mock())
        assert resource_manager != HTTPResourceManager("bar", client)

    @pytest.mark.parametrize(
        ["response", "expected_result", "exception"],
        [
            pytest.param(MagicMock(json=MagicMock(return_value={})), {}, None, id="ok"),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="", request=MagicMock(), response=MagicMock(status_code=http.HTTPStatus.BAD_REQUEST)
                        )
                    )
                ),
                None,
                exceptions.IntegrityError(resource="foo"),
                id="bad_request",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="",
                            request=MagicMock(),
                            response=MagicMock(status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR),
                        )
                    )
                ),
                None,
                httpx.HTTPStatusError,
                id="internal_server_error",
            ),
        ],
        indirect=["exception"],
    )
    async def test_create(self, resource_manager, response, expected_result, exception):
        data = {"foo": "bar"}

        with exception, patch.object(resource_manager._client, "post", return_value=response):
            result = await resource_manager.create(data)
            assert resource_manager._client.post.await_args_list == [call("foo/", json=data)]
            assert result == expected_result

    @pytest.mark.parametrize(
        ["response", "expected_result", "exception"],
        [
            pytest.param(
                MagicMock(json=MagicMock(return_value={})),
                {},
                None,
                id="ok",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="", request=MagicMock(), response=MagicMock(status_code=http.HTTPStatus.NOT_FOUND)
                        )
                    )
                ),
                None,
                exceptions.NotFoundError(resource="foo", id="bar"),
                id="not_found",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="",
                            request=MagicMock(),
                            response=MagicMock(status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR),
                        )
                    )
                ),
                None,
                httpx.HTTPStatusError,
                id="internal_server_error",
            ),
        ],
        indirect=["exception"],
    )
    async def test_retrieve(self, resource_manager, response, expected_result, exception):
        with exception, patch.object(resource_manager._client, "get", return_value=response):
            result = await resource_manager.retrieve("bar")
            assert resource_manager._client.get.await_args_list == [call("foo/bar/")]
            assert result == expected_result

    @pytest.mark.parametrize(
        ["response", "expected_result", "exception"],
        [
            pytest.param(
                MagicMock(json=MagicMock(return_value={})),
                {},
                None,
                id="ok",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="", request=MagicMock(), response=MagicMock(status_code=http.HTTPStatus.NOT_FOUND)
                        )
                    )
                ),
                None,
                exceptions.NotFoundError(resource="foo", id="bar"),
                id="not_found",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="", request=MagicMock(), response=MagicMock(status_code=http.HTTPStatus.BAD_REQUEST)
                        )
                    )
                ),
                None,
                exceptions.IntegrityError(resource="foo"),
                id="bad_request",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="",
                            request=MagicMock(),
                            response=MagicMock(status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR),
                        )
                    )
                ),
                None,
                httpx.HTTPStatusError,
                id="internal_server_error",
            ),
        ],
        indirect=["exception"],
    )
    async def test_update(self, resource_manager, response, expected_result, exception):
        data = {"foo": "bar"}

        with exception, patch.object(resource_manager._client, "put", return_value=response):
            result = await resource_manager.update("bar", data=data)
            assert resource_manager._client.put.await_args_list == [call("foo/bar/", json=data)]
            assert result == expected_result

    @pytest.mark.parametrize(
        ["response", "expected_result", "exception"],
        [
            pytest.param(
                MagicMock(json=MagicMock(return_value={})),
                {},
                None,
                id="ok",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="", request=MagicMock(), response=MagicMock(status_code=http.HTTPStatus.NOT_FOUND)
                        )
                    )
                ),
                None,
                exceptions.NotFoundError(resource="foo", id="bar"),
                id="not_found",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="", request=MagicMock(), response=MagicMock(status_code=http.HTTPStatus.BAD_REQUEST)
                        )
                    )
                ),
                None,
                exceptions.IntegrityError(resource="foo"),
                id="bad_request",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="",
                            request=MagicMock(),
                            response=MagicMock(status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR),
                        )
                    )
                ),
                None,
                httpx.HTTPStatusError,
                id="internal_server_error",
            ),
        ],
        indirect=["exception"],
    )
    async def test_partial_update(self, resource_manager, response, expected_result, exception):
        data = {"foo": "bar"}

        with exception, patch.object(resource_manager._client, "patch", return_value=response):
            result = await resource_manager.partial_update("bar", data=data)
            assert resource_manager._client.patch.await_args_list == [call("foo/bar/", json=data)]
            assert result == expected_result

    @pytest.mark.parametrize(
        ["response", "expected_result", "exception"],
        [
            pytest.param(
                MagicMock(),
                None,
                None,
                id="ok",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="", request=MagicMock(), response=MagicMock(status_code=http.HTTPStatus.NOT_FOUND)
                        )
                    )
                ),
                None,
                exceptions.NotFoundError(resource="foo", id="bar"),
                id="not_found",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="",
                            request=MagicMock(),
                            response=MagicMock(status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR),
                        )
                    )
                ),
                None,
                httpx.HTTPStatusError,
                id="internal_server_error",
            ),
        ],
        indirect=["exception"],
    )
    async def test_delete(self, resource_manager, response, expected_result, exception):
        with exception, patch.object(resource_manager._client, "delete", return_value=response):
            result = await resource_manager.delete("bar")
            assert resource_manager._client.delete.await_args_list == [call("foo/bar/")]
            assert result == expected_result

    @pytest.mark.parametrize(
        ["pagination", "response", "expected_result", "expected_calls"],
        [
            pytest.param(
                "page_number",
                [
                    MagicMock(json=MagicMock(return_value={"data": [{"foo": 1}]})),
                    MagicMock(json=MagicMock(return_value={"data": [{"bar": 2}]})),
                    MagicMock(json=MagicMock(return_value={"data": []})),
                ],
                [{"foo": 1}, {"bar": 2}],
                [
                    call("foo/", params={"page_number": 1}),
                    call("foo/", params={"page_number": 2}),
                    call("foo/", params={"page_number": 3}),
                ],
                id="ok_page_number",
            ),
            pytest.param(
                "limit_offset",
                [
                    MagicMock(json=MagicMock(return_value={"data": [{"foo": 1}]})),
                    MagicMock(json=MagicMock(return_value={"data": [{"bar": 2}]})),
                    MagicMock(json=MagicMock(return_value={"data": []})),
                ],
                [{"foo": 1}, {"bar": 2}],
                [
                    call("foo/", params={"offset": 0}),
                    call("foo/", params={"offset": 1}),
                    call("foo/", params={"offset": 2}),
                ],
                id="ok_limit_offset",
            ),
        ],
    )
    async def test_list(self, resource_manager, pagination, response, expected_result, expected_calls):
        with patch.object(resource_manager._client, "get", side_effect=response):
            assert [e async for e in resource_manager.list(pagination=pagination)] == expected_result
            assert resource_manager._client.get.await_args_list == expected_calls

    @pytest.mark.parametrize(
        ["response", "expected_result", "exception"],
        [
            pytest.param(MagicMock(json=MagicMock(return_value=[{}, {}])), [{}, {}], None, id="ok"),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="", request=MagicMock(), response=MagicMock(status_code=http.HTTPStatus.BAD_REQUEST)
                        )
                    )
                ),
                None,
                exceptions.IntegrityError(resource="foo"),
                id="bad_request",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="",
                            request=MagicMock(),
                            response=MagicMock(status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR),
                        )
                    )
                ),
                None,
                httpx.HTTPStatusError,
                id="internal_server_error",
            ),
        ],
        indirect=["exception"],
    )
    async def test_replace(self, resource_manager, response, expected_result, exception):
        data = [{"foo": "bar"}, {"bar": "foo"}]
        with exception, patch.object(resource_manager._client, "put", return_value=response):
            result = await resource_manager.replace(data)
            assert resource_manager._client.put.await_args_list == [call("foo/", json=data)]
            assert result == expected_result

    @pytest.mark.parametrize(
        ["response", "expected_result", "exception"],
        [
            pytest.param(MagicMock(json=MagicMock(return_value=[{}, {}])), [{}, {}], None, id="ok"),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="", request=MagicMock(), response=MagicMock(status_code=http.HTTPStatus.BAD_REQUEST)
                        )
                    )
                ),
                None,
                exceptions.IntegrityError(resource="foo"),
                id="bad_request",
            ),
            pytest.param(
                MagicMock(
                    raise_for_status=MagicMock(
                        side_effect=httpx.HTTPStatusError(
                            message="",
                            request=MagicMock(),
                            response=MagicMock(status_code=http.HTTPStatus.INTERNAL_SERVER_ERROR),
                        )
                    )
                ),
                None,
                httpx.HTTPStatusError,
                id="internal_server_error",
            ),
        ],
        indirect=["exception"],
    )
    async def test_partial_replace(self, resource_manager, response, expected_result, exception):
        data = [{"foo": "bar"}, {"bar": "foo"}]
        with exception, patch.object(resource_manager._client, "patch", return_value=response):
            result = await resource_manager.partial_replace(data)
            assert resource_manager._client.patch.await_args_list == [call("foo/", json=data)]
            assert result == expected_result

    @pytest.mark.parametrize(
        ["response", "expected_result", "exception"],
        [
            pytest.param(MagicMock(json=MagicMock(return_value={"deleted": 1})), 1, None, id="ok"),
        ],
        indirect=["exception"],
    )
    async def test_drop(self, resource_manager, response, expected_result, exception):
        with exception, patch.object(resource_manager._client, "delete", return_value=response):
            result = await resource_manager.drop()
            assert resource_manager._client.delete.await_args_list == [call("foo/")]
            assert result == expected_result


class TestCaseHTTPResourceRepository:
    @pytest.fixture(scope="function")
    def client(self):
        return Mock(spec=Client)

    @pytest.fixture(scope="function")
    def resource_manager(self):
        return Mock(spec=HTTPResourceManager)

    @pytest.fixture(scope="function")
    def repository(self, client, resource_manager):
        class Repository(HTTPResourceRepository):
            _resource = "foo"

        r = Repository(client)
        with patch.object(r, "_resource_manager", resource_manager):
            yield r

    async def test_init(self, client):
        class Repository(HTTPResourceRepository):
            _resource = "foo"

        repository = Repository(client)
        assert repository._client == client
        assert repository._resource_manager == HTTPResourceManager("foo", client)

    async def test_eq(self, client):
        class Repository(HTTPResourceRepository):
            _resource = "foo"

        repository = Repository(client)
        assert repository == Repository(client)
        assert repository != Repository(Mock(spec=Client))

    async def test_create(self, repository, resource_manager):
        data = {"foo": "bar"}

        await repository.create(data)

        assert resource_manager.create.call_args_list == [call(data)]

    async def test_retrieve(self, repository, resource_manager):
        id = uuid.uuid4()

        await repository.retrieve(id)

        assert resource_manager.retrieve.call_args_list == [call(id)]

    async def test_update(self, repository, resource_manager):
        id = uuid.uuid4()
        data = {"foo": "bar"}

        await repository.update(id, data)

        assert resource_manager.update.call_args_list == [call(id, data)]

    async def test_partial_update(self, repository, resource_manager):
        id = uuid.uuid4()
        data = {"foo": "bar"}

        await repository.partial_update(id, data)

        assert resource_manager.partial_update.call_args_list == [call(id, data)]

    async def test_delete(self, repository, resource_manager):
        id = uuid.uuid4()

        await repository.delete(id)

        assert resource_manager.delete.call_args_list == [call(id)]

    async def test_list(self, repository, resource_manager):
        repository.list(pagination="page_number")

        assert resource_manager.list.call_args_list == [call(pagination="page_number")]

    async def test_replace(self, repository, resource_manager):
        data = [{"foo": "bar"}]

        await repository.replace(data)

        assert resource_manager.replace.call_args_list == [call(data)]

    async def test_partial_replace(self, repository, resource_manager):
        data = [{"foo": "bar"}]

        await repository.partial_replace(data)

        assert resource_manager.partial_replace.call_args_list == [call(data)]

    async def test_drop(self, repository, resource_manager):
        await repository.drop()

        assert resource_manager.drop.call_args_list == [call()]



import datetime
import json
import tempfile
import uuid

import pytest

import flama
from flama.serialize.data_structures import Compression
from flama.serialize.types import Framework


class TestCaseSerialize:
    @pytest.fixture(scope="function")
    def artifact(self):
        with tempfile.NamedTemporaryFile(mode="w", suffix=".json") as tmp:
            json.dump({"foo": "bar"}, tmp)
            yield tmp.name

    @pytest.mark.parametrize(
        ("lib", "model", "serialized_model_class"),
        (
            pytest.param(Framework.sklearn, "sklearn", "sklearn", id="sklearn"),
            pytest.param(Framework.sklearn, "sklearn-pipeline", "sklearn-pipeline", id="sklearn-pipeline"),
            pytest.param(Framework.tensorflow, "tensorflow", "tensorflow", id="tensorflow"),
            pytest.param(Framework.torch, "torch", "torch", id="torch"),
            # TODO: Add keras
        ),
        indirect=["model", "serialized_model_class"],
    )
    def test_serialize(self, lib, artifact, model, serialized_model_class):
        id_ = uuid.uuid4()
        timestamp = datetime.datetime.utcnow()
        params = {"param": "1"}
        metrics = {"metric": "1"}
        extra = {"foo": "bar"}

        with tempfile.NamedTemporaryFile(suffix=".flm") as tmp:
            flama.dump(
                model,
                tmp.name,
                compression=Compression.fast,
                model_id=id_,
                timestamp=timestamp,
                params=params,
                metrics=metrics,
                extra=extra,
                artifacts={"foo.json": artifact},
            )

            load_model = flama.load(tmp.name)

        assert isinstance(load_model.model, serialized_model_class)
        assert load_model.meta.id == id_
        assert load_model.meta.timestamp == timestamp
        assert load_model.meta.framework.lib == lib
        assert load_model.meta.model.params == params
        assert load_model.meta.model.metrics == metrics
        assert load_model.meta.extra == extra
        assert "foo.json" in load_model.artifacts



import datetime
import typing as t
import uuid

from flama import exceptions
from flama.ddd.repositories.sqlalchemy import SQLAlchemyTableRepository
from flama.resources import data_structures
from flama.resources.exceptions import ResourceAttributeError
from flama.resources.resource import Resource, ResourceType

try:
    import sqlalchemy
    from sqlalchemy.dialects import postgresql
except Exception:  # pragma: no cover
    raise exceptions.DependencyNotInstalled(
        dependency=exceptions.DependencyNotInstalled.Dependency.sqlalchemy, dependant=__name__
    )

__all__ = ["RESTResource", "RESTResourceType"]


class RESTResourceType(ResourceType):
    def __new__(mcs, name: str, bases: tuple[type], namespace: dict[str, t.Any]):
        """Resource metaclass for defining basic behavior for REST resources:
        * Create _meta attribute containing some metadata (model, schemas...).
        * Adds methods related to REST resource (create, retrieve, update, delete...) listed in METHODS class attribute.

        :param name: Class name.
        :param bases: List of superclasses.
        :param namespace: Variables namespace used to create the class.
        """
        if not mcs._is_abstract(namespace):
            try:
                # Get model
                model = mcs._get_model(bases, namespace)
                namespace["model"] = model.table

                # Get input and output schemas
                resource_schemas = mcs._get_schemas(name, bases, namespace)
                namespace["schemas"] = resource_schemas
            except AttributeError as e:
                raise ResourceAttributeError(str(e), name)

            namespace.setdefault("_meta", data_structures.Metadata()).namespaces.update(
                {
                    "rest": {"model": model, "schemas": resource_schemas},
                    "ddd": {
                        "repository": type(f"{name}Repository", (SQLAlchemyTableRepository,), {"_table": model.table})
                    },
                }
            )

        return super().__new__(mcs, name, bases, namespace)

    @staticmethod
    def _is_abstract(namespace: dict[str, t.Any]) -> bool:
        return namespace.get("__module__") == "flama.resources.rest" and namespace.get("__qualname__") == "RESTResource"

    @classmethod
    def _get_model(cls, bases: t.Sequence[t.Any], namespace: dict[str, t.Any]) -> data_structures.Model:
        """Look for the resource model and checks if a primary key is defined with a valid type.

        :param bases: List of superclasses.
        :param namespace: Variables namespace used to create the class.
        :return: Resource model.
        """
        model = cls._get_attribute("model", bases, namespace, metadata_namespace="rest")

        # Already defined model probably because resource inheritance, so no need to create it
        if isinstance(model, data_structures.Model):
            return model

        # Resource define model as a sqlalchemy Table, so extract necessary info from it
        elif isinstance(model, sqlalchemy.Table):
            # Get model primary key
            model_pk_columns = list(sqlalchemy.inspect(model).primary_key.columns.values())

            # Check primary key exists and is a single column
            if len(model_pk_columns) != 1:
                raise AttributeError(ResourceAttributeError.PK_NOT_FOUND)

            model_pk = model_pk_columns[0]
            model_pk_name = model_pk.name

            # Check primary key is a valid type
            try:
                model_pk_mapping: dict[type, type] = {
                    sqlalchemy.Integer: int,
                    sqlalchemy.String: str,
                    sqlalchemy.Date: datetime.date,
                    sqlalchemy.DateTime: datetime.datetime,
                    postgresql.UUID: uuid.UUID,
                }
                model_pk_type = model_pk_mapping[model_pk.type.__class__]
            except KeyError:
                raise AttributeError(ResourceAttributeError.PK_WRONG_TYPE)

            return data_structures.Model(
                table=model, primary_key=data_structures.PrimaryKey(model_pk_name, model_pk_type)
            )

        raise AttributeError(ResourceAttributeError.MODEL_INVALID)

    @classmethod
    def _get_schemas(cls, name: str, bases: t.Sequence[t.Any], namespace: dict[str, t.Any]) -> data_structures.Schemas:
        """Look for the resource schema or the pair of input and output schemas.

        :param name: Class name.
        :param bases: List of superclasses.
        :param namespace: Variables namespace used to create the class.
        :return: Resource schemas.
        """
        try:
            return data_structures.Schemas(
                input=data_structures.Schema(
                    name="Input" + name,
                    schema=cls._get_attribute("input_schema", bases, namespace, metadata_namespace="rest"),
                ),
                output=data_structures.Schema(
                    name="Output" + name,
                    schema=cls._get_attribute("output_schema", bases, namespace, metadata_namespace="rest"),
                ),
            )
        except AttributeError:
            ...

        try:
            schema = data_structures.Schema(
                name=name, schema=cls._get_attribute("schema", bases, namespace, metadata_namespace="rest")
            )
            return data_structures.Schemas(input=schema, output=schema)
        except AttributeError:
            ...

        try:
            schemas: data_structures.Schemas = cls._get_attribute(
                "schemas", bases, namespace, metadata_namespace="rest"
            )
            return schemas
        except AttributeError:
            ...

        raise AttributeError(ResourceAttributeError.SCHEMA_NOT_FOUND)


class RESTResource(Resource, metaclass=RESTResourceType):
    model: sqlalchemy.Table
    schema: t.Any
    input_schema: t.Any
    output_schema: t.Any



import builtins
import http
import typing as t
import uuid

import httpx

from flama.ddd import exceptions
from flama.ddd.repositories import BaseRepository

if t.TYPE_CHECKING:
    from flama.client import Client

__all__ = ["HTTPRepository", "HTTPResourceManager", "HTTPResourceRepository"]


class HTTPRepository(BaseRepository):
    def __init__(self, client: "Client", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._client = client

    def __eq__(self, other):
        return isinstance(other, HTTPRepository) and self._client == other._client


class HTTPResourceManager:
    def __init__(self, resource: str, client: "Client"):
        self._client = client
        self.resource = resource.rstrip("/")

    def __eq__(self, other):
        return (
            isinstance(other, HTTPResourceManager) and self._client == other._client and self.resource == other.resource
        )

    async def create(self, data: dict[str, t.Any]) -> dict[str, t.Any]:
        """Create a new element in the collection.

        :param data: The data to create the element.
        :return: The element created.
        :raises IntegrityError: If the resource already exists or cannot be created.
        """
        try:
            response = await self._client.post(f"{self.resource}/", json=data)
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == http.HTTPStatus.BAD_REQUEST:
                raise exceptions.IntegrityError(resource=self.resource)
            raise

        return response.json()

    async def retrieve(self, id: t.Union[str, uuid.UUID]) -> dict[str, t.Any]:
        """Retrieve an element from the collection.

        :param id: The id of the element.
        :return: The element retrieved.
        :raises NotFoundError: If the resource is not found.
        """
        try:
            response = await self._client.get(f"{self.resource}/{id}/")
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == http.HTTPStatus.NOT_FOUND:
                raise exceptions.NotFoundError(resource=self.resource, id=id)
            raise

        return response.json()

    async def update(self, id: t.Union[str, uuid.UUID], data: dict[str, t.Any]) -> dict[str, t.Any]:
        """Update an element in the collection.

        :param id: The id of the element.
        :param data: The data to update the element.
        :return: The element updated.
        :raises NotFoundError: If the resource is not found.
        :raises IntegrityError: If wrong input data.
        """
        try:
            response = await self._client.put(f"{self.resource}/{id}/", json=data)
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == http.HTTPStatus.NOT_FOUND:
                raise exceptions.NotFoundError(resource=self.resource, id=id)
            if e.response.status_code == http.HTTPStatus.BAD_REQUEST:
                raise exceptions.IntegrityError(resource=self.resource)
            raise
        return response.json()

    async def partial_update(self, id: t.Union[str, uuid.UUID], data: dict[str, t.Any]) -> dict[str, t.Any]:
        """Partially update an element in the collection.

        :param id: The id of the element.
        :param data: The data to update the element.
        :return: The element updated.
        :raises NotFoundError: If the resource is not found.
        :raises IntegrityError: If wrong input data.
        """
        try:
            response = await self._client.patch(f"{self.resource}/{id}/", json=data)
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == http.HTTPStatus.NOT_FOUND:
                raise exceptions.NotFoundError(resource=self.resource, id=id)
            if e.response.status_code == http.HTTPStatus.BAD_REQUEST:
                raise exceptions.IntegrityError(resource=self.resource)
            raise
        return response.json()

    async def delete(self, id: t.Union[str, uuid.UUID]) -> None:
        """Delete an element from the collection.

        :param id: The id of the element.
        :raises NotFoundError: If the resource is not found.
        """
        try:
            response = await self._client.delete(f"{self.resource}/{id}/")
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == http.HTTPStatus.NOT_FOUND:
                raise exceptions.NotFoundError(resource=self.resource, id=id)
            raise

    async def _fetch_page_elements(self, **params: t.Any) -> t.AsyncIterator[dict[str, t.Any]]:
        """Fetch elements of the collection.

        :param params: The parameters to use in the request.
        :return: Async iterator of the elements.
        :raises StopIteration: If there are no more elements to fetch.
        """
        response = await self._client.get(f"{self.resource}/", params=params)

        data = response.json()["data"]
        if not data:
            raise exceptions.Empty()

        for element in data:
            yield element

    async def _page_number_paginated(self) -> t.AsyncIterable[dict[str, t.Any]]:
        """Fetch elements of the collection paginated by page number.

        :return: Async iterable of the elements.
        """
        page_number = 0
        while True:
            try:
                page_number += 1
                async for element in self._fetch_page_elements(page_number=page_number):
                    yield element
            except exceptions.Empty:
                break

    async def _limit_offset_paginated(self) -> t.AsyncIterable[dict[str, t.Any]]:
        """Fetch elements of the collection paginated by limit and offset.

        :return: Async iterable of the elements.
        """
        offset = 0
        while True:
            try:
                async for element in self._fetch_page_elements(offset=offset):
                    offset += 1
                    yield element
            except exceptions.Empty:
                break

    async def list(self, *, pagination: str = "page_number") -> t.AsyncIterable[dict[str, t.Any]]:
        """List all the elements in the collection.

        :param pagination: The pagination technique.
        :return: Async iterable of the elements.
        """

        iterator = self._page_number_paginated() if pagination == "page_number" else self._limit_offset_paginated()

        async for element in iterator:
            yield element

    async def replace(self, data: builtins.list[dict[str, t.Any]]) -> builtins.list[dict[str, t.Any]]:
        """Replace elements in the collection.

        :param data: The data to replace the elements.
        :return: The elements replaced.
        """
        try:
            response = await self._client.put(f"{self.resource}/", json=data)
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == http.HTTPStatus.BAD_REQUEST:
                raise exceptions.IntegrityError(resource=self.resource)
            raise

        return [element for element in response.json()]

    async def partial_replace(self, data: builtins.list[dict[str, t.Any]]) -> builtins.list[dict[str, t.Any]]:
        """Partially replace elements in the collection.

        :param data: The data to replace the elements.
        :return: The elements replaced.
        """
        try:
            response = await self._client.patch(f"{self.resource}/", json=data)
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code == http.HTTPStatus.BAD_REQUEST:
                raise exceptions.IntegrityError(resource=self.resource)
            raise

        return [element for element in response.json()]

    async def drop(self) -> int:
        """Drop the collection.

        :return: The number of elements deleted.
        """
        response = await self._client.delete(f"{self.resource}/")
        response.raise_for_status()
        return response.json()["deleted"]


class HTTPResourceRepository(HTTPRepository):
    """Base class for HTTP repositories. It provides a client to make requests to the API.

    The `_resource` attribute must be defined in the subclasses to specify the resource to use in the requests.
    """

    _resource: str

    def __init__(self, client: "Client"):
        """Initialise the repository.

        :param client: The client to use to make the requests.
        """
        super().__init__(client)
        self._resource_manager = HTTPResourceManager(self._resource, client)

    async def create(self, data: dict[str, t.Any]) -> dict[str, t.Any]:
        """Create a new element in the collection.

        :param data: The data to create the element.
        :return: The element created.
        """
        return await self._resource_manager.create(data)

    async def retrieve(self, id: uuid.UUID) -> dict[str, t.Any]:
        """Retrieve an element from the collection.

        :param id: The id of the element.
        :return: The element retrieved.
        """
        return await self._resource_manager.retrieve(id)

    async def update(self, id: uuid.UUID, data: dict[str, t.Any]) -> dict[str, t.Any]:
        """Update an element in the collection.

        :param id: The id of the element.
        :param data: The data to update the element.
        :return: The element updated.
        """
        return await self._resource_manager.update(id, data)

    async def partial_update(self, id: uuid.UUID, data: dict[str, t.Any]) -> dict[str, t.Any]:
        """Partially update an element in the collection.

        :param id: The id of the element.
        :param data: The data to update the element.
        :return: The element updated.
        """
        return await self._resource_manager.partial_update(id, data)

    async def delete(self, id: uuid.UUID) -> None:
        """Delete an element from the collection.

        :param id: The id of the element.
        """
        return await self._resource_manager.delete(id)

    def list(self, *, pagination: str = "page_number") -> t.AsyncIterable[dict[str, t.Any]]:
        """List all the elements in the collection.

        :param pagination: The pagination technique.
        :return: Async iterable of the elements.
        """
        return self._resource_manager.list(pagination=pagination)

    async def replace(self, data: builtins.list[dict[str, t.Any]]) -> builtins.list[dict[str, t.Any]]:
        """Replace elements in the collection.

        :param data: The data to replace the elements.
        :return: The elements replaced.
        """
        return await self._resource_manager.replace(data)

    async def partial_replace(self, data: builtins.list[dict[str, t.Any]]) -> builtins.list[dict[str, t.Any]]:
        """Partially replace elements in the collection.

        :param data: The data to replace the elements.
        :return: The elements replaced.
        """
        return await self._resource_manager.partial_replace(data)

    async def drop(self) -> int:
        """Drop the collection.

        :return: The number of elements deleted.
        """
        return await self._resource_manager.drop()



import importlib.util
import sys
import typing as t

import sqlalchemy

from flama import Flama


def installed(module: str) -> bool:
    """Check if a module is installed.

    :param module: module name.
    :return: True if installed and importable.
    """
    return module in sys.modules or importlib.util.find_spec(module) is not None


class ExceptionContext:
    def __init__(self, context, exception: t.Optional[Exception] = None):
        self.context = context
        self.exception = exception

    def __enter__(self):
        return self.context.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.context.__exit__(exc_type, exc_val, exc_tb)

    def __bool__(self) -> bool:
        return self.exception is not None


class SQLAlchemyContext:
    def __init__(self, app: Flama, tables: list[sqlalchemy.Table]):
        self.app = app
        self.tables = tables

    async def __aenter__(self):
        async with self.app.sqlalchemy.engine.begin() as connection:
            await connection.run_sync(self.app.sqlalchemy.metadata.create_all, tables=self.tables)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        async with self.app.sqlalchemy.engine.begin() as connection:
            await connection.run_sync(self.app.sqlalchemy.metadata.drop_all, tables=self.tables)



import functools
import typing as t

import starlette.middleware.authentication
import starlette.middleware.base
import starlette.middleware.cors
import starlette.middleware.gzip
import starlette.middleware.httpsredirect
import starlette.middleware.trustedhost

from flama import concurrency
from flama.debug.middleware import ExceptionMiddleware, ServerErrorMiddleware

if t.TYPE_CHECKING:
    from flama import Flama, types
    from flama.http import Request, Response

__all__ = [
    "BaseHTTPMiddleware",
    "CORSMiddleware",
    "ExceptionMiddleware",
    "GZipMiddleware",
    "HTTPSRedirectMiddleware",
    "Middleware",
    "MiddlewareStack",
    "SessionMiddleware",
    "TrustedHostMiddleware",
]

try:
    import starlette.middleware.sessions

    class SessionMiddleware(starlette.middleware.sessions.SessionMiddleware):
        def __init__(self, app: "types.App", *args, **kwargs):
            super().__init__(app, *args, **kwargs)  # type: ignore[arg-type]

        async def __call__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:  # type: ignore[overrid]
            return await super().__call__(scope, receive, send)  # type: ignore[assignment]

except ModuleNotFoundError:
    SessionMiddleware = None  # type: ignore[assignment]


class BaseHTTPMiddleware(starlette.middleware.base.BaseHTTPMiddleware):
    def __init__(self, app: "types.App", *args, **kwargs):
        super().__init__(app, *args, **kwargs)  # type: ignore[arg-type]

    async def __call__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:  # type: ignore[overrid]
        return await super().__call__(scope, receive, send)  # type: ignore[assignment]


class CORSMiddleware(starlette.middleware.cors.CORSMiddleware):
    def __init__(self, app: "types.App", *args, **kwargs):
        super().__init__(app, *args, **kwargs)  # type: ignore[arg-type]

    async def __call__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:  # type: ignore[overrid]
        return await super().__call__(scope, receive, send)  # type: ignore[assignment]


class GZipMiddleware(starlette.middleware.gzip.GZipMiddleware):
    def __init__(self, app: "types.App", *args, **kwargs):
        super().__init__(app, *args, **kwargs)  # type: ignore[arg-type]

    async def __call__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:  # type: ignore[overrid]
        return await super().__call__(scope, receive, send)  # type: ignore[assignment]


class HTTPSRedirectMiddleware(starlette.middleware.httpsredirect.HTTPSRedirectMiddleware):
    def __init__(self, app: "types.App", *args, **kwargs):
        super().__init__(app, *args, **kwargs)  # type: ignore[arg-type]

    async def __call__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:  # type: ignore[overrid]
        return await super().__call__(scope, receive, send)  # type: ignore[assignment]


class TrustedHostMiddleware(starlette.middleware.trustedhost.TrustedHostMiddleware):
    def __init__(self, app: "types.App", *args, **kwargs):
        super().__init__(app, *args, **kwargs)  # type: ignore[arg-type]

    async def __call__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:  # type: ignore[overrid]
        return await super().__call__(scope, receive, send)  # type: ignore[assignment]


class Middleware:
    def __init__(self, middleware: "types.Middleware", **kwargs: t.Any) -> None:
        self.middleware = middleware
        self.kwargs = kwargs

    def __call__(self, app: "types.App") -> t.Union["types.MiddlewareClass", "types.App"]:
        return self.middleware(app, **self.kwargs)

    def __repr__(self) -> str:
        name = self.__class__.__name__
        middleware_name = (
            self.middleware.__class__.__name__
            if isinstance(self.middleware, types.MiddlewareClass)
            else self.middleware.__name__
        )
        args = ", ".join([middleware_name] + [f"{key}={value!r}" for key, value in self.kwargs.items()])
        return f"{name}({args})"


class MiddlewareStack:
    def __init__(self, app: "Flama", middleware: t.Sequence[Middleware], debug: bool):
        self.app = app
        self.middleware = list(reversed(middleware))
        self.debug = debug
        self._exception_handlers: dict[t.Union[int, type[Exception]], t.Callable[[Request, Exception], Response]] = {}
        self._stack: t.Optional[t.Union[types.MiddlewareClass, types.App]] = None

    @property
    def stack(
        self,
    ) -> t.Union["types.MiddlewareClass", "types.App"]:
        if self._stack is None:
            self._stack = functools.reduce(
                lambda app, middleware: middleware(app=app),
                [
                    Middleware(ExceptionMiddleware, handlers=self._exception_handlers, debug=self.debug),
                    *self.middleware,
                    Middleware(ServerErrorMiddleware, debug=self.debug),
                ],
                self.app.router,
            )

        return self._stack

    @stack.deleter
    def stack(self):
        self._stack = None

    def add_exception_handler(
        self, key: t.Union[int, type[Exception]], handler: t.Callable[["Request", Exception], "Response"]
    ):
        """Adds a new handler for an exception type or a HTTP status code.

        :param key: Exception type or HTTP status code.
        :param handler: Exception handler.
        """
        self._exception_handlers[key] = handler
        del self.stack

    def add_middleware(self, middleware: Middleware):
        """Adds a new middleware to the stack.

        :param middleware: Middleware.
        """
        self.middleware.append(middleware)
        del self.stack

    async def __call__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:
        await concurrency.run(self.stack, scope, receive, send)



import logging
import typing as t

from flama import exceptions, types, url
from flama.injection import Component, Components
from flama.lifespan import Lifespan
from flama.routing.routes.base import BaseRoute
from flama.routing.routes.http import Route
from flama.routing.routes.mount import Mount
from flama.routing.routes.websocket import WebSocketRoute

if t.TYPE_CHECKING:
    from flama.applications import Flama, types


__all__ = ["Router"]

logger = logging.getLogger(__name__)


class Router:
    def __init__(
        self,
        routes: t.Optional[t.Sequence[BaseRoute]] = None,
        *,
        components: t.Optional[t.Union[t.Sequence["Component"], set["Component"]]] = None,
        lifespan: t.Optional[t.Callable[[t.Optional["Flama"]], t.AsyncContextManager]] = None,
        root: t.Optional["Flama"] = None,
    ):
        """A router for containing all routes and mount points.

        :param routes: Routes part of this router.
        :param components: Components registered in this router.
        :param lifespan: Lifespan function.
        :param root: Flama application.
        """
        self.routes = [] if routes is None else list(routes)
        self.components = Components(components if components else set())
        self.lifespan = Lifespan(lifespan)

        if root:
            self.build(root)

    def __eq__(self, other: t.Any) -> bool:
        return isinstance(other, Router) and self.routes == other.routes

    async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
        logger.debug("Request: %s", str(scope))
        if scope["type"] not in ("http", "websocket", "lifespan"):
            raise ValueError(f"Wrong scope type ({scope['type']})")

        if scope["type"] == "lifespan":
            await self.lifespan(scope, receive, send)
            return

        if "router" not in scope:
            scope["router"] = self

        route, route_scope = self.resolve_route(scope)
        await route(route_scope, receive, send)

    def build(self, app: "Flama") -> None:
        """Build step for routes.

        Just build the parameters' descriptor part of RouteParametersMixin.

        :param app: Flama app.
        """
        for route in self.routes:
            route.build(app)

    def add_component(self, component: Component):
        """Register a new component.

        :param component: Component to register.
        """
        self.components = Components(self.components + Components([component]))

    def add_route(
        self,
        path: t.Optional[str] = None,
        endpoint: t.Optional[types.HTTPHandler] = None,
        methods: t.Optional[list[str]] = None,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        *,
        route: t.Optional[Route] = None,
        root: t.Optional["Flama"] = None,
        pagination: t.Optional[types.Pagination] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> Route:
        """Register a new HTTP route in this router under given path.

        :param path: URL path.
        :param endpoint: HTTP endpoint.
        :param methods: List of valid HTTP methods (only applies for routes).
        :param name: Endpoint or route name.
        :param include_in_schema: True if this route or endpoint should be declared as part of the API schema.
        :param route: HTTP route.
        :param root: Flama application.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the route or endpoint.
        :return: Route.
        """
        if path is not None and endpoint is not None:
            route = Route(
                path,
                endpoint=endpoint,
                methods=methods,
                name=name,
                include_in_schema=include_in_schema,
                pagination=pagination,
                tags=tags,
            )

        if route is None:
            raise exceptions.ApplicationError("Either 'path' and 'endpoint' or 'route' variables are needed")

        self.routes.append(route)

        route.build(root)

        return route

    def route(
        self,
        path: str,
        methods: t.Optional[list[str]] = None,
        name: t.Optional[str] = None,
        include_in_schema: bool = True,
        *,
        root: t.Optional["Flama"] = None,
        pagination: t.Optional[types.Pagination] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> t.Callable[[types.HTTPHandler], types.HTTPHandler]:
        """Decorator version for registering a new HTTP route in this router under given path.

        :param path: URL path.
        :param methods: List of valid HTTP methods (only applies for routes).
        :param name: Endpoint or route name.
        :param include_in_schema: True if this route or endpoint should be declared as part of the API schema.
        :param root: Flama application.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the endpoint.
        :return: Decorated route.
        """

        def decorator(func: types.HTTPHandler) -> types.HTTPHandler:
            self.add_route(
                path,
                func,
                methods=methods,
                name=name,
                include_in_schema=include_in_schema,
                root=root,
                pagination=pagination,
                tags=tags,
            )
            return func

        return decorator

    def add_websocket_route(
        self,
        path: t.Optional[str] = None,
        endpoint: t.Optional[types.WebSocketHandler] = None,
        name: t.Optional[str] = None,
        *,
        route: t.Optional[WebSocketRoute] = None,
        root: t.Optional["Flama"] = None,
        pagination: t.Optional[types.Pagination] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> WebSocketRoute:
        """Register a new websocket route in this router under given path.

        :param path: URL path.
        :param endpoint: Websocket function or endpoint.
        :param name: Websocket route name.
        :param route: Specific route class.
        :param root: Flama application.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the websocket route.
        :return: Websocket route.
        """
        if path is not None and endpoint is not None:
            route = WebSocketRoute(path, endpoint=endpoint, name=name, pagination=pagination, tags=tags)

        if route is None:
            raise exceptions.ApplicationError("Either 'path' and 'endpoint' or 'route' variables are needed")

        self.routes.append(route)

        route.build(root)

        return route

    def websocket_route(
        self,
        path: str,
        name: t.Optional[str] = None,
        *,
        root: t.Optional["Flama"] = None,
        pagination: t.Optional[types.Pagination] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> t.Callable[[types.WebSocketHandler], types.WebSocketHandler]:
        """Decorator version for registering a new websocket route in this router under given path.

        :param path: URL path.
        :param name: Websocket route name.
        :param root: Flama application.
        :param pagination: Apply a pagination technique.
        :param tags: Tags to add to the websocket route.
        :return: Decorated websocket route.
        """

        def decorator(func: types.WebSocketHandler) -> types.WebSocketHandler:
            self.add_websocket_route(path, func, name=name, root=root, pagination=pagination, tags=tags)
            return func

        return decorator

    def mount(
        self,
        path: t.Optional[str] = None,
        app: t.Optional[types.App] = None,
        name: t.Optional[str] = None,
        *,
        mount: t.Optional[Mount] = None,
        root: t.Optional["Flama"] = None,
        tags: t.Optional[dict[str, t.Any]] = None,
    ) -> Mount:
        """Register a new mount point containing an ASGI app in this router under given path.

        :param path: URL path.
        :param app: ASGI app to mount.
        :param name: Route name.
        :param mount: Mount.
        :param root: Flama application.
        :param tags: Tags to add to the mount.
        :return: Mount.
        """
        if path is not None and app is not None:
            mount = Mount(path, app=app, name=name, tags=tags)

        if mount is None:
            raise exceptions.ApplicationError("Either 'path' and 'app' or 'mount' variables are needed")

        self.routes.append(mount)

        mount.build(root)

        return mount

    def resolve_route(self, scope: types.Scope) -> tuple[BaseRoute, types.Scope]:
        """Look for a route that matches given ASGI scope.

        :param scope: ASGI scope.
        :raise MethodNotAllowedException: If route is resolved but http method is not valid.
        :raise NotFoundException: If route cannot be resolved.
        :return: Route and its scope.
        """
        partial = None
        partial_allowed_methods: set[str] = set()

        for route in self.routes:
            m = route.match(scope)
            if m == route.Match.full:
                route_scope = types.Scope({**scope, **route.route_scope(scope)})

                if isinstance(route, Mount):
                    try:
                        return route.app.resolve_route(route_scope)  # type: ignore[no-any-return,union-attr]
                    except AttributeError:
                        ...

                return route, route_scope
            elif m == route.Match.partial:
                partial = route
                if isinstance(route, Route):
                    partial_allowed_methods |= route.methods

        if partial:
            route_scope = types.Scope({**scope, **partial.route_scope(scope)})
            raise exceptions.MethodNotAllowedException(
                route_scope.get("root_path", "") + route_scope["path"], route_scope["method"], partial_allowed_methods
            )

        raise exceptions.NotFoundException(
            path=scope.get("root_path", "") + scope["path"], params=scope.get("path_params")
        )

    def resolve_url(self, name: str, **path_params: t.Any) -> url.URL:
        """Look for a route URL given the route name and path params.

        :param name: Route name.
        :param path_params: Path params.
        :return: Route URL.
        """
        for route in self.routes:
            try:
                return route.resolve_url(name, **path_params)
            except exceptions.NotFoundException:
                pass
        raise exceptions.NotFoundException(params=path_params, name=name)



import functools
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

from flama import exceptions, routing, types
from flama.applications import Flama
from flama.url import Path


class TestCaseBaseEndpointWrapper:
    @pytest.fixture(scope="function")
    def wrapper_cls(self):
        class _Wrapper(routing.BaseEndpointWrapper):
            async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None: ...

        return _Wrapper

    @pytest.fixture(scope="function")
    def handler(self):
        async def foo(): ...

        return foo

    @pytest.fixture(scope="function")
    def endpoint(self, wrapper_cls, handler):
        return wrapper_cls(handler)

    @pytest.mark.parametrize(
        ["pagination", "exception"],
        [
            pytest.param(None, None, id="ok"),
            pytest.param("wrong", KeyError("wrong"), id="wrong_pagination"),
        ],
        indirect=["exception"],
    )
    def test_init(self, wrapper_cls, handler, pagination, exception):
        with exception:
            endpoint = wrapper_cls(handler, pagination=pagination)

            assert endpoint.handler == handler

    def test_get(self, endpoint):
        endpoint_ = endpoint

        class Foo:
            endpoint = endpoint_

        assert isinstance(Foo.endpoint, functools.partial)

    def test_eq(self, wrapper_cls, handler):
        async def bar(): ...

        endpoint = wrapper_cls(handler)
        endpoint_foo = wrapper_cls(handler)
        endpoint_bar = wrapper_cls(bar)

        assert endpoint == endpoint_foo
        assert endpoint != endpoint_bar
        assert endpoint_foo != endpoint_bar


class TestCaseBaseRoute:
    @pytest.fixture(scope="function")
    def wrapper_cls(self):
        class _Wrapper(routing.BaseEndpointWrapper):
            async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None: ...

        return _Wrapper

    @pytest.fixture(scope="function")
    def route_cls(self):
        class _Route(routing.BaseRoute):
            async def __call__(self, scope: types.Scope, receive: types.Receive, send: types.Send) -> None:
                return await self.handle(scope, receive, send)

        return _Route

    @pytest.fixture(scope="function")
    def route(self, route_cls, wrapper_cls):
        return route_cls("/", wrapper_cls(lambda: None), name="foo", include_in_schema=False)

    def test_init(self, route_cls):
        def foo(): ...

        app = foo
        route = route_cls(
            "/",
            app,
            name="foo",
            include_in_schema=False,
            tags={"tag": "tag", "list_tag": ["foo", "bar"], "dict_tag": {"foo": "bar"}},
        )

        assert route.path == "/"
        assert route.app == app
        assert route.endpoint == foo
        assert route.name == "foo"
        assert route.include_in_schema is False
        assert route.tags == {"tag": "tag", "list_tag": ["foo", "bar"], "dict_tag": {"foo": "bar"}}

    def test_eq(self, route_cls):
        def foo(): ...

        assert route_cls("/", foo, name="foo") == route_cls("/", foo, name="foo")
        assert route_cls("/", foo, name="foo") != route_cls("/", foo, name="bar")

    def test_repr(self, route_cls):
        def foo(): ...

        assert repr(route_cls("/", foo, name="foo")) == "_Route(path='/', name='foo')"

    @pytest.mark.parametrize(["app"], (pytest.param(MagicMock(spec=Flama), id="app"), pytest.param(None, id="no_app")))
    def test_build(self, app, route):
        expected_calls = [call(app)] if app else []
        with patch.object(route, "parameters") as parameters_mock:
            route.build(app)

        assert parameters_mock.build.call_args_list == expected_calls

    def test_endpoint_handlers(self, route):
        assert route.endpoint_handlers() == {}

    async def test_handle(self, route_cls, asgi_scope, asgi_receive, asgi_send):
        app = AsyncMock()
        route = route_cls("/", app)

        await route.handle(asgi_scope, asgi_receive, asgi_send)

        assert app.call_args_list == [call(asgi_scope, asgi_receive, asgi_send)]

    @pytest.mark.parametrize(
        ["path_match_return", "result"],
        (
            pytest.param(MagicMock(match=Path.Match.exact), routing.BaseRoute.Match.full, id="match"),
            pytest.param(MagicMock(match=Path.Match.none), routing.BaseRoute.Match.none, id="no_match"),
        ),
    )
    def test_match(self, path_match_return, result, route, asgi_scope):
        with patch.object(route.path, "match", return_value=path_match_return):
            assert route.match(asgi_scope) == result

    def test_route_scope(self, route, asgi_scope):
        assert route.route_scope(asgi_scope) == types.Scope({})

    @pytest.mark.parametrize(
        ["name", "params", "exception"],
        (
            pytest.param(
                "foo",
                {"bar": 1},
                None,
                id="found",
            ),
            pytest.param(
                "bar",
                {},
                exceptions.NotFoundException(name="bar"),
                id="not_found_wrong_name",
            ),
            pytest.param(
                "foo",
                {"wrong": 1},
                exceptions.NotFoundException(params={"wrong": 1}, name="foo"),
                id="not_found_wrong_params",
            ),
            pytest.param(
                "foo",
                {"bar": 1, "wrong": 1},
                exceptions.NotFoundException(params={"bar": 1, "wrong": 1}, name="foo"),
                id="error_remaining_params",
            ),
        ),
        indirect=["exception"],
    )
    def test_resolve_url(self, route_cls, name, params, exception):
        route = route_cls("/foo/{bar:int}", lambda bar: None, name="foo")

        with exception:
            assert route.resolve_url(name=name, **params)



import dataclasses
import datetime
import enum
import html
import importlib.util
import inspect
import json
import os
import pathlib
import typing as t
import uuid
import warnings

import jinja2
import starlette.requests
import starlette.responses
import starlette.schemas

from flama import compat, exceptions, schemas, types, url

__all__ = [
    "Method",
    "Request",
    "Response",
    "HTMLResponse",
    "PlainTextResponse",
    "JSONResponse",
    "RedirectResponse",
    "StreamingResponse",
    "FileResponse",
    "APIResponse",
    "APIErrorResponse",
    "HTMLFileResponse",
    "HTMLTemplatesEnvironment",
    "HTMLTemplateResponse",
    "OpenAPIResponse",
]

Method = compat.StrEnum(  # PORT: Replace compat when stop supporting 3.10
    "Method", ["GET", "HEAD", "POST", "PUT", "DELETE", "CONNECT", "OPTIONS", "TRACE", "PATCH"]
)
Request = starlette.requests.Request


class Response(starlette.responses.Response):
    async def __call__(  # type: ignore[override]
        self, scope: types.Scope, receive: types.Receive, send: types.Send
    ) -> None:
        await super().__call__(scope, receive, send)  # type: ignore[arg-type]

    def __eq__(self, value: object, /) -> bool:
        return (
            isinstance(value, Response)
            and self.status_code == value.status_code
            and getattr(self, "media_type") == getattr(value, "media_type")
            and self.background == value.background
            and self.body == value.body
            and self.headers == value.headers
        )


class HTMLResponse(starlette.responses.HTMLResponse, Response):
    async def __call__(  # type: ignore[override]
        self, scope: types.Scope, receive: types.Receive, send: types.Send
    ) -> None:
        await super().__call__(scope, receive, send)  # type: ignore[arg-type]


class PlainTextResponse(starlette.responses.PlainTextResponse, Response):
    async def __call__(  # type: ignore[override]
        self, scope: types.Scope, receive: types.Receive, send: types.Send
    ) -> None:
        await super().__call__(scope, receive, send)  # type: ignore[arg-type]


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, (pathlib.Path, os.PathLike, uuid.UUID, url.Path, url.URL)):
            return str(o)
        if isinstance(o, (bytes, bytearray)):
            return o.decode("utf-8")
        if isinstance(o, enum.Enum):
            return o.value
        if isinstance(o, (set, frozenset)):
            return list(o)
        if isinstance(o, (datetime.datetime, datetime.date, datetime.time)):
            return o.isoformat()
        if isinstance(o, datetime.timedelta):
            # split seconds to larger units
            seconds = o.total_seconds()
            minutes, seconds = divmod(seconds, 60)
            hours, minutes = divmod(minutes, 60)
            days, hours = divmod(hours, 24)
            days, hours, minutes = map(int, (days, hours, minutes))
            seconds = round(seconds, 6)

            formatted_units = (
                (days, f"{days:02d}".lstrip("0") + "D"),
                (hours, f"{hours:02d}".lstrip("0") + "H"),
                (minutes, f"{minutes:02d}".lstrip("0") + "M"),
                (seconds, f"{seconds:.6f}".strip("0") + "S"),
            )

            return "P" + "".join([formatted_value for value, formatted_value in formatted_units if value])
        if inspect.isclass(o) and issubclass(o, BaseException):
            return o.__name__
        if isinstance(o, BaseException):
            return repr(o)
        if dataclasses.is_dataclass(o) and not isinstance(o, type):
            return dataclasses.asdict(o)
        return super().default(o)


class JSONResponse(starlette.responses.JSONResponse, Response):
    async def __call__(  # type: ignore[override]
        self, scope: types.Scope, receive: types.Receive, send: types.Send
    ) -> None:
        await super().__call__(scope, receive, send)  # type: ignore[arg-type]

    def render(self, content: t.Any) -> bytes:
        return json.dumps(
            content, ensure_ascii=False, allow_nan=False, indent=None, separators=(",", ":"), cls=EnhancedJSONEncoder
        ).encode("utf-8")


class RedirectResponse(starlette.responses.RedirectResponse, Response):
    async def __call__(  # type: ignore[override]
        self, scope: types.Scope, receive: types.Receive, send: types.Send
    ) -> None:
        await super().__call__(scope, receive, send)  # type: ignore[arg-type]


class StreamingResponse(starlette.responses.StreamingResponse, Response):
    async def __call__(  # type: ignore[override]
        self, scope: types.Scope, receive: types.Receive, send: types.Send
    ) -> None:
        await super().__call__(scope, receive, send)  # type: ignore[arg-type]


class FileResponse(starlette.responses.FileResponse, Response):
    async def __call__(  # type: ignore[override]
        self, scope: types.Scope, receive: types.Receive, send: types.Send
    ) -> None:
        await super().__call__(scope, receive, send)  # type: ignore[arg-type]


class APIResponse(JSONResponse):
    media_type = "application/json"

    def __init__(self, content: t.Any = None, schema: t.Any = None, *args, **kwargs):
        self.schema = schema
        super().__init__(content, *args, **kwargs)

    def render(self, content: t.Any):
        if self.schema is not None:
            try:
                content = schemas.Schema.from_type(self.schema).dump(content)
            except schemas.SchemaValidationError as e:
                raise exceptions.SerializationError(status_code=500, detail=e.errors)

        if not content:
            return b""

        return super().render(content)


class APIErrorResponse(APIResponse):
    def __init__(
        self,
        detail: t.Any,
        status_code: int = 400,
        exception: t.Optional[Exception] = None,
        headers: t.Optional[dict[str, str]] = None,
        *args,
        **kwargs,
    ):
        content = {
            "detail": detail,
            "error": str(exception.__class__.__name__) if exception is not None else None,
            "status_code": status_code,
            "headers": headers,
        }

        super().__init__(
            content,
            schema=t.Annotated[schemas.Schema, schemas.SchemaMetadata(schemas.schemas.APIError)],
            status_code=status_code,
            *args,
            **kwargs,
        )

        self.detail = detail
        self.exception = exception


class HTMLFileResponse(HTMLResponse):
    def __init__(self, path: str, *args, **kwargs):
        try:
            with open(path) as f:
                content = f.read()
        except Exception as e:
            raise exceptions.HTTPException(status_code=500, detail=str(e))

        super().__init__(content, *args, **kwargs)


class HTMLTemplatesEnvironment(jinja2.Environment):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            **{
                **kwargs,
                "comment_start_string": "||*",
                "comment_end_string": "*||",
                "block_start_string": "||%",
                "block_end_string": "%||",
                "variable_start_string": "||@",
                "variable_end_string": "@||",
            },
        )

        self.filters["safe"] = self.safe
        self.filters["safe_json"] = self.safe_json

    @t.overload
    def _escape(self, value: str) -> str: ...
    @t.overload
    def _escape(self, value: bool) -> bool: ...
    @t.overload
    def _escape(self, value: int) -> int: ...
    @t.overload
    def _escape(self, value: float) -> float: ...
    @t.overload
    def _escape(self, value: None) -> None: ...
    @t.overload
    def _escape(self, value: types.JSONField) -> types.JSONField: ...
    def _escape(self, value: types.JSONField) -> types.JSONField:
        if isinstance(value, (list, tuple)):
            return [self._escape(x) for x in value]

        if isinstance(value, dict):
            return {k: self._escape(v) for k, v in value.items()}

        if isinstance(value, str):
            return html.escape(value).replace("\n", "&#10;&#13;")

        return value

    def safe(self, value: str) -> str:
        return self._escape(value)

    def safe_json(self, value: types.JSONField):
        return json.dumps(self._escape(value)).replace('"', '\\"')


class HTMLTemplateResponse(HTMLResponse):
    templates = HTMLTemplatesEnvironment(loader=jinja2.FileSystemLoader(pathlib.Path(os.curdir) / "templates"))

    def __init__(self, template: str, context: t.Optional[dict[str, t.Any]] = None, *args, **kwargs):
        if context is None:
            context = {}

        super().__init__(self.templates.get_template(template).render(**context), *args, **kwargs)


class _FlamaLoader(jinja2.PackageLoader):
    def __init__(self):
        spec = importlib.util.find_spec("flama")
        if spec is None or spec.origin is None:
            raise exceptions.ApplicationError("Flama package not found")

        templates_path = pathlib.Path(spec.origin).parent.joinpath("templates")
        if not templates_path.exists():
            warnings.warn("Templates folder not found in the Flama package")
            templates_path.mkdir(exist_ok=True)

        super().__init__(package_name="flama", package_path="templates")


class _FlamaTemplateResponse(HTMLTemplateResponse):
    templates = HTMLTemplatesEnvironment(loader=_FlamaLoader())


class OpenAPIResponse(starlette.schemas.OpenAPIResponse, Response):
    async def __call__(  # type: ignore[override]
        self, scope: types.Scope, receive: types.Receive, send: types.Send
    ) -> None:
        await super().__call__(scope, receive, send)  # type: ignore[arg-type]

    def render(self, content: t.Any) -> bytes:
        if not isinstance(content, dict):
            raise ValueError("The schema must be a dictionary")

        return json.dumps(content).encode("utf-8")



import inspect
import typing as t

import pydantic
from pydantic.fields import FieldInfo
from pydantic.json_schema import model_json_schema

from flama import compat
from flama.injection import Parameter
from flama.schemas.adapter import Adapter
from flama.schemas.exceptions import SchemaGenerationError, SchemaValidationError
from flama.types import JSONSchema

__all__ = ["PydanticAdapter"]

Schema = pydantic.BaseModel
Field = FieldInfo


class PydanticAdapter(Adapter[Schema, Field]):
    def build_field(
        self,
        name: str,
        type_: type,
        nullable: bool = False,
        required: bool = True,
        default: t.Any = None,
        multiple: bool = False,
        **kwargs,
    ) -> Field:
        if not required:
            kwargs["default"] = None if default is Parameter.empty else default

        annotation: t.Any = type_

        if multiple:
            annotation = list[annotation]

        if nullable:
            annotation = t.Optional[annotation]

        if default is Parameter.empty:
            field = FieldInfo.from_annotation(annotation)
        else:
            field = FieldInfo.from_annotated_attribute(annotation, default)

        return field

    def build_schema(
        self,
        *,
        name: t.Optional[str] = None,
        schema: t.Optional[t.Union[Schema, type[Schema]]] = None,
        fields: t.Optional[dict[str, type[Field]]] = None,
        partial: bool = False,
    ) -> type[Schema]:
        fields_ = {
            **{
                name: (field.annotation, field)
                for name, field in (self.unique_schema(schema).model_fields.items() if self.is_schema(schema) else {})
            },
            **{name: (field.annotation, field) for name, field in (fields.items() if fields else {})},
        }

        if partial:
            for name, (annotation, field) in fields_.items():
                field.default = None
                fields_[name] = (t.Optional[annotation], field)

        return pydantic.create_model(name or self.DEFAULT_SCHEMA_NAME, **fields_)

    def validate(
        self, schema: t.Union[Schema, type[Schema]], values: dict[str, t.Any], *, partial: bool = False
    ) -> dict[str, t.Any]:
        schema_cls = self.unique_schema(schema)

        try:
            return schema_cls(**values).model_dump(exclude_unset=partial)
        except pydantic.ValidationError as errors:
            raise SchemaValidationError(
                errors={
                    ".".join(str(x) for x in error.get("loc", [])): error for error in errors.errors(include_url=False)
                }
            )

    def load(self, schema: t.Union[Schema, type[Schema]], value: dict[str, t.Any]) -> Schema:
        schema_cls = self.unique_schema(schema)

        return schema_cls(**value)

    def dump(self, schema: t.Union[Schema, type[Schema]], value: dict[str, t.Any]) -> dict[str, t.Any]:
        schema_cls = self.unique_schema(schema)

        return self.validate(schema_cls, value)

    def name(self, schema: t.Union[Schema, type[Schema]], *, prefix: t.Optional[str] = None) -> str:
        s = self.unique_schema(schema)
        schema_name = f"{prefix or ''}{s.__qualname__}"
        return schema_name if s.__module__ == "builtins" else f"{s.__module__}.{schema_name}"

    def to_json_schema(self, schema: t.Union[type[Schema], type[Field]]) -> JSONSchema:
        try:
            if self.is_schema(schema):
                json_schema = model_json_schema(schema, ref_template="#/components/schemas/{model}")
                if "$defs" in json_schema:
                    del json_schema["$defs"]
            elif self.is_field(schema):
                json_schema = model_json_schema(
                    self.build_schema(fields={"x": schema}), ref_template="#/components/schemas/{model}"
                )["properties"]["x"]
                if not schema.title:  # Pydantic is introducing a default title, so we drop it
                    del json_schema["title"]
            else:
                raise TypeError("Not a valid schema class or field")

            return json_schema
        except Exception as e:
            raise SchemaGenerationError from e

    def unique_schema(self, schema: t.Union[Schema, type[Schema]]) -> type[Schema]:
        return schema.__class__ if isinstance(schema, Schema) else schema

    def _get_field_type(self, field: Field) -> t.Any:
        if not self.is_field(field):
            return field

        if t.get_origin(field.annotation) == list:
            return self._get_field_type(t.get_args(field.annotation)[0])

        if t.get_origin(field.annotation) == dict:
            return self._get_field_type(t.get_args(field.annotation)[1])

        return field.annotation

    def schema_fields(
        self, schema: type[Schema]
    ) -> dict[str, tuple[t.Union[type, list[type], dict[str, type]], Field]]:
        return {name: (self._get_field_type(field), field) for name, field in schema.model_fields.items()}

    def is_schema(self, obj: t.Any) -> compat.TypeGuard[type[Schema]]:  # PORT: Replace compat when stop supporting 3.9
        if t.get_origin(obj):
            obj = t.get_origin(obj)

        return inspect.isclass(obj) and issubclass(obj, Schema)

    def is_field(self, obj: t.Any) -> compat.TypeGuard[type[Field]]:  # PORT: Replace compat when stop supporting 3.9
        return isinstance(obj, Field)



from unittest.mock import Mock

import pytest

from flama.applications import Flama
from flama.resources.crud import CRUDResource
from flama.resources.routing import ResourceRoute
from flama.sqlalchemy import SQLAlchemyModule


class TestCaseResourcesModule:
    @pytest.fixture(scope="function")
    def tags(self):
        return {
            "create": {"tag": "create"},
            "retrieve": {"tag": "retrieve"},
            "update": {"tag": "update"},
            "partial_update": {"tag": "partial-update"},
            "delete": {"tag": "delete"},
            "list": {"tag": "list"},
            "replace": {"tag": "replace"},
            "partial_replace": {"tag": "partial-replace"},
            "drop": {"tag": "drop"},
        }

    @pytest.fixture(scope="function")
    def app(self):
        return Flama(schema=None, docs=None, modules={SQLAlchemyModule("sqlite+aiosqlite://")})

    def test_add_resource(self, app, puppy_model, puppy_schema, tags):
        class PuppyResource(CRUDResource):
            name = "puppy"
            model = puppy_model
            schema = puppy_schema

        resource = PuppyResource()
        app.resources.add_resource("/puppy/", resource, tags=tags)

        try:
            assert len(app.routes) == 1
            assert isinstance(app.routes[0], ResourceRoute)
            resource_route = app.routes[0]
            assert [(route.path, route.methods, route.endpoint, route.tags) for route in resource_route.routes] == [
                ("/", {"POST"}, resource.create, {"tag": "create"}),
                ("/{resource_id}/", {"GET", "HEAD"}, resource.retrieve, {"tag": "retrieve"}),
                ("/{resource_id}/", {"PUT"}, resource.update, {"tag": "update"}),
                ("/{resource_id}/", {"PATCH"}, resource.partial_update, {"tag": "partial-update"}),
                ("/{resource_id}/", {"DELETE"}, resource.delete, {"tag": "delete"}),
                ("/", {"GET", "HEAD"}, resource.list, {"tag": "list"}),
                ("/", {"PUT"}, resource.replace, {"tag": "replace"}),
                ("/", {"PATCH"}, resource.partial_replace, {"tag": "partial-replace"}),
                ("/", {"DELETE"}, resource.drop, {"tag": "drop"}),
            ]
            assert PuppyResource._meta.name in app.resources.worker._resources_repositories.registered
        finally:
            app.resources.remove_repository(PuppyResource._meta.name)

    def test_add_resource_decorator(self, app, puppy_model, puppy_schema, tags):
        class PuppyResource(CRUDResource):
            name = "puppy"
            model = puppy_model
            schema = puppy_schema

        resource = app.resources.resource("/puppy/", tags=tags)(
            PuppyResource()
        )  # Apply decoration to an instance in order to check endpoints

        try:
            assert len(app.routes) == 1
            assert isinstance(app.routes[0], ResourceRoute)
            resource_route = app.routes[0]
            assert [(route.path, route.methods, route.endpoint, route.tags) for route in resource_route.routes] == [
                ("/", {"POST"}, resource.create, {"tag": "create"}),
                ("/{resource_id}/", {"GET", "HEAD"}, resource.retrieve, {"tag": "retrieve"}),
                ("/{resource_id}/", {"PUT"}, resource.update, {"tag": "update"}),
                ("/{resource_id}/", {"PATCH"}, resource.partial_update, {"tag": "partial-update"}),
                ("/{resource_id}/", {"DELETE"}, resource.delete, {"tag": "delete"}),
                ("/", {"GET", "HEAD"}, resource.list, {"tag": "list"}),
                ("/", {"PUT"}, resource.replace, {"tag": "replace"}),
                ("/", {"PATCH"}, resource.partial_replace, {"tag": "partial-replace"}),
                ("/", {"DELETE"}, resource.drop, {"tag": "drop"}),
            ]
            assert PuppyResource._meta.name in app.resources.worker._resources_repositories.registered
        finally:
            app.resources.remove_repository(PuppyResource._meta.name)

    def test_add_resource_wrong(self, app):
        with pytest.raises(ValueError, match=""):
            app.resources.add_resource("/puppy/", Mock())

    def test_mount_resource_declarative(self, puppy_model, puppy_schema, tags):
        class PuppyResource(CRUDResource):
            name = "puppy"
            model = puppy_model
            schema = puppy_schema

        route = ResourceRoute("/puppy/", PuppyResource, tags=tags)

        app = Flama(routes=[route], schema=None, docs=None)

        try:
            assert len(app.router.routes) == 1
            assert len(app.routes) == 1
            assert isinstance(app.routes[0], ResourceRoute)
            resource_route = app.routes[0]
            assert [(route.path, route.methods, route.endpoint, route.tags) for route in resource_route.routes] == [
                ("/", {"POST"}, resource_route.resource.create, {"tag": "create"}),
                ("/{resource_id}/", {"GET", "HEAD"}, resource_route.resource.retrieve, {"tag": "retrieve"}),
                ("/{resource_id}/", {"PUT"}, resource_route.resource.update, {"tag": "update"}),
                ("/{resource_id}/", {"PATCH"}, resource_route.resource.partial_update, {"tag": "partial-update"}),
                ("/{resource_id}/", {"DELETE"}, resource_route.resource.delete, {"tag": "delete"}),
                ("/", {"GET", "HEAD"}, resource_route.resource.list, {"tag": "list"}),
                ("/", {"PUT"}, resource_route.resource.replace, {"tag": "replace"}),
                ("/", {"PATCH"}, resource_route.resource.partial_replace, {"tag": "partial-replace"}),
                ("/", {"DELETE"}, resource_route.resource.drop, {"tag": "drop"}),
            ]
            assert isinstance(resource_route.resource, PuppyResource)
            assert PuppyResource._meta.name in app.resources.worker._resources_repositories.registered
        finally:
            app.resources.remove_repository(PuppyResource._meta.name)



import datetime
import http
import uuid
from unittest.mock import MagicMock, patch

import pytest

from flama import Flama, authentication
from flama.exceptions import HTTPException
from flama.telemetry.data_structures import Authentication, Endpoint, Error, Request, Response, TelemetryData

TOKEN = (
    "eyJhbGciOiAiSFMyNTYiLCAidHlwIjogIkpXVCJ9.eyJkYXRhIjogeyJmb28iOiAiYmFyIn0sICJpYXQiOiAwfQ==.J3zdedMZSFNOimstjJat0V"
    "28rM_b1UU62XCp9dg_5kg="
)


@pytest.fixture(scope="function")
def secret():
    return uuid.UUID(int=0)


@pytest.fixture(scope="function")
def app(secret):
    return Flama(
        schema=None,
        docs=None,
        components=[
            authentication.AccessTokenComponent(secret=secret.bytes),
            authentication.RefreshTokenComponent(secret=secret.bytes),
        ],
    )


@pytest.fixture(scope="function", autouse=True)
def add_routes(app):
    @app.route("/")
    def root(): ...

    @app.route("/query_parameters/")
    def query_parameters(x: int):
        return {"x": x}

    @app.route("/path_parameters/{x:int}/")
    def path_parameters(x: int):
        return {"x": x}

    @app.route("/tags/", tags={"foo": "bar"})
    def tags(): ...

    @app.route("/name/", name="foo")
    def name(): ...


@pytest.fixture(scope="function")
def asgi_scope(app, asgi_scope):
    asgi_scope["app"] = app
    return asgi_scope


class TestCaseAuthentication:
    @pytest.mark.parametrize(
        ["scope", "result"],
        [
            pytest.param(
                {"path": "/"},
                {"access": None, "refresh": None},
                id="no_auth",
            ),
            pytest.param(
                {"path": "/", "headers": [(b"cookie", f"access_token={TOKEN}".encode())]},
                {
                    "access": {
                        "header": {"alg": "HS256", "typ": "JWT"},
                        "payload": {"data": {"foo": "bar"}, "iat": 0},
                    },
                    "refresh": None,
                },
                id="access",
            ),
            pytest.param(
                {"path": "/", "headers": [(b"cookie", f"refresh_token={TOKEN}".encode())]},
                {
                    "access": None,
                    "refresh": {
                        "header": {"alg": "HS256", "typ": "JWT"},
                        "payload": {"data": {"foo": "bar"}, "iat": 0},
                    },
                },
                id="refresh",
            ),
            pytest.param(
                {
                    "path": "/",
                    "headers": [(b"cookie", f"access_token={TOKEN}; refresh_token={TOKEN}".encode())],
                },
                {
                    "access": {
                        "header": {"alg": "HS256", "typ": "JWT"},
                        "payload": {"data": {"foo": "bar"}, "iat": 0},
                    },
                    "refresh": {
                        "header": {"alg": "HS256", "typ": "JWT"},
                        "payload": {"data": {"foo": "bar"}, "iat": 0},
                    },
                },
                id="access_and_refresh",
            ),
        ],
    )
    async def test_from_scope(self, asgi_scope, asgi_receive, asgi_send, scope, result):
        asgi_scope.update(scope)
        now = datetime.datetime.now()
        with patch("datetime.datetime", MagicMock(now=MagicMock(return_value=now))):
            data = await Authentication.from_scope(scope=asgi_scope, receive=asgi_receive, send=asgi_send)

            assert data.to_dict() == result


class TestCaseEndpoint:
    @pytest.mark.parametrize(
        ["scope", "result"],
        [
            pytest.param(
                {"path": "/"},
                {"name": "root", "path": "/", "tags": {}},
                id="default_name",
            ),
            pytest.param(
                {"path": "/name/"},
                {"name": "foo", "path": "/name/", "tags": {}},
                id="explicit_name",
            ),
            pytest.param(
                {"path": "/tags/"},
                {"name": "tags", "path": "/tags/", "tags": {"foo": "bar"}},
                id="tags",
            ),
        ],
    )
    async def test_from_scope(self, asgi_scope, asgi_receive, asgi_send, scope, result):
        asgi_scope.update(scope)
        now = datetime.datetime.now()
        with patch("datetime.datetime", MagicMock(now=MagicMock(return_value=now))):
            data = await Endpoint.from_scope(scope=asgi_scope, receive=asgi_receive, send=asgi_send)

            assert data.to_dict() == result


class TestCaseRequest:
    @pytest.mark.parametrize(
        ["scope", "result"],
        [
            pytest.param(
                {"path": "/"},
                {
                    "body": b"",
                    "cookies": {},
                    "headers": {},
                    "path_parameters": {},
                    "query_parameters": {},
                },
                id="empty",
            ),
            pytest.param(
                {"path": "/path_parameters/1/"},
                {
                    "body": b"",
                    "cookies": {},
                    "headers": {},
                    "path_parameters": {"x": 1},
                    "query_parameters": {},
                },
                id="path_parameters",
            ),
            pytest.param(
                {"path": "/query_parameters/", "query_string": b"x=1"},
                {
                    "body": b"",
                    "cookies": {},
                    "headers": {},
                    "path_parameters": {},
                    "query_parameters": {"x": "1"},
                },
                id="query_parameters",
            ),
            pytest.param(
                {"path": "/", "headers": [(b"foo", b"bar")]},
                {
                    "body": b"",
                    "cookies": {},
                    "headers": {"foo": "bar"},
                    "path_parameters": {},
                    "query_parameters": {},
                },
                id="headers",
            ),
            pytest.param(
                {"path": "/", "headers": [(b"cookie", b"foo=bar")]},
                {
                    "body": b"",
                    "cookies": {
                        "foo": {
                            "comment": "",
                            "domain": "",
                            "expires": "",
                            "httponly": "",
                            "max-age": "",
                            "path": "",
                            "samesite": "",
                            "secure": "",
                            "value": "bar",
                            "version": "",
                        }
                    },
                    "headers": {"cookie": "foo=bar"},
                    "path_parameters": {},
                    "query_parameters": {},
                },
                id="cookies",
            ),
        ],
    )
    async def test_from_scope(self, asgi_scope, asgi_receive, asgi_send, scope, result):
        asgi_scope.update(scope)
        now = datetime.datetime.now()
        result["timestamp"] = now.isoformat()
        with patch("datetime.datetime", MagicMock(now=MagicMock(return_value=now))):
            data = await Request.from_scope(scope=asgi_scope, receive=asgi_receive, send=asgi_send)

            assert data.to_dict() == result


class TestCaseResponse:
    @pytest.mark.parametrize(
        ["headers", "body", "status_code", "result"],
        [
            pytest.param(
                {},
                b"",
                None,
                {
                    "body": b"",
                    "cookies": {},
                    "headers": {},
                    "status_code": None,
                },
                id="empty",
            ),
            pytest.param(
                {},
                b"foo",
                http.HTTPStatus.OK,
                {
                    "body": b"foo",
                    "cookies": {},
                    "headers": {},
                    "status_code": http.HTTPStatus.OK,
                },
                id="body",
            ),
            pytest.param(
                {"foo": "bar"},
                b"",
                http.HTTPStatus.OK,
                {
                    "body": b"",
                    "cookies": {},
                    "headers": {"foo": "bar"},
                    "status_code": http.HTTPStatus.OK,
                },
                id="headers",
            ),
            pytest.param(
                {"cookie": "foo=bar"},
                b"",
                http.HTTPStatus.OK,
                {
                    "body": b"",
                    "cookies": {
                        "foo": {
                            "comment": "",
                            "domain": "",
                            "expires": "",
                            "httponly": "",
                            "max-age": "",
                            "path": "",
                            "samesite": "",
                            "secure": "",
                            "value": "bar",
                            "version": "",
                        }
                    },
                    "headers": {"cookie": "foo=bar"},
                    "status_code": http.HTTPStatus.OK,
                },
                id="cookies",
            ),
        ],
    )
    def test_init(self, headers, body, status_code, result):
        now = datetime.datetime.now()
        result["timestamp"] = now.isoformat()
        with patch("datetime.datetime", MagicMock(now=MagicMock(return_value=now))):
            data = Response(headers=headers, body=body, status_code=status_code)

            assert data.to_dict() == result


class TestCaseError:
    @pytest.mark.parametrize(
        ["exception", "result"],
        [
            pytest.param(
                ValueError("foo"),
                {"detail": "foo", "status_code": None},
                id="exception",
            ),
            pytest.param(
                HTTPException(status_code=400, detail="foo"),
                {"detail": "foo", "status_code": 400},
                id="http",
            ),
        ],
    )
    async def test_from_exception(self, exception, result):
        now = datetime.datetime.now()
        result["timestamp"] = now.isoformat()
        with patch("datetime.datetime", MagicMock(now=MagicMock(return_value=now))):
            try:
                raise exception
            except Exception as e:
                error = await Error.from_exception(exception=e)

            assert error.to_dict() == result


class TestCaseTelemetryData:
    async def test_from_scope(self, asgi_scope, asgi_receive, asgi_send):
        now = datetime.datetime.now()
        with patch("datetime.datetime", MagicMock(now=MagicMock(return_value=now))):
            data = await TelemetryData.from_scope(scope=asgi_scope, receive=asgi_receive, send=asgi_send)

            assert data.to_dict() == {
                "authentication": {"access": None, "refresh": None},
                "endpoint": {"name": "root", "path": "/", "tags": {}},
                "error": None,
                "extra": {},
                "request": {
                    "body": b"",
                    "cookies": {},
                    "headers": {},
                    "path_parameters": {},
                    "query_parameters": {},
                    "timestamp": now.isoformat(),
                },
                "response": None,
                "type": "http",
            }



import importlib.util
import sys
import typing as t
from types import ModuleType

from flama.exceptions import DependencyNotInstalled
from flama.schemas.data_structures import Field, Parameter, Schema
from flama.schemas.exceptions import SchemaParseError, SchemaValidationError
from flama.schemas.types import (
    SchemaMetadata,
    SchemaType,
    get_schema_metadata,
    is_schema,
    is_schema_multiple,
    is_schema_partial,
)

if t.TYPE_CHECKING:
    from flama.schemas.adapter import Adapter

__all__ = [
    "SchemaValidationError",
    "SchemaParseError",
    "Field",
    "Schema",
    "adapter",
    "fields",
    "lib",
    "schemas",
    "SchemaMetadata",
    "SchemaType",
    "get_schema_metadata",
    "is_schema",
    "is_schema_multiple",
    "is_schema_partial",
]

adapter: "Adapter"
fields: dict[t.Any, "Parameter"] = {}
lib: t.Optional[ModuleType] = None
schemas: t.Any = None


class Module:
    SCHEMA_LIBS = ("pydantic", "typesystem", "marshmallow")

    def __init__(self) -> None:
        self.lib: ModuleType

    @property
    def installed(self) -> list[str]:
        return [x for x in self.SCHEMA_LIBS if x in sys.modules or importlib.util.find_spec(x) is not None]

    @property
    def available(self) -> t.Generator[str, None, None]:
        for library in self.installed:
            try:
                importlib.import_module(f"flama.schemas._libs.{library}")
                yield library
            except ModuleNotFoundError:
                pass

    def setup(self, library: t.Optional[str] = None):
        try:
            if library is None:
                library = next(self.available)
        except StopIteration:
            raise DependencyNotInstalled(
                msg="No schema library is installed. Install one of your preference following instructions from: "
                "https://flama.dev/docs/getting-started/installation#extras"
            )
        self.lib = importlib.import_module(f"flama.schemas._libs.{library}")

        global schemas, lib, fields, adapter, Field, Schema
        schemas = self.lib.schemas
        lib = self.lib.lib
        fields = self.lib.fields
        adapter = self.lib.adapter


# Find the first schema lib available and setup the module using it
_module = Module()
_module.setup()

# Check that at least one of the schema libs is installed
if _module.lib is None:
    raise DependencyNotInstalled(
        msg=f"Any of the schema libraries ({', '.join(_module.SCHEMA_LIBS)}) must be installed."
    )



import typing as t
from unittest.mock import MagicMock, call, patch

import pytest

from flama import Flama, endpoints, types


class TestCaseBaseEndpoint:
    @pytest.fixture(scope="function")
    def app(self, route):
        app = MagicMock(spec=Flama)
        app.router = MagicMock(resolve_route=MagicMock(side_effect=lambda x: (route, x)))
        return app

    @pytest.fixture(scope="function")
    def route(self):
        return MagicMock()

    @pytest.fixture
    def endpoint(self):
        class _Endpoint(endpoints.BaseEndpoint):
            @classmethod
            def allowed_handlers(cls) -> dict[str, t.Callable]: ...

            async def dispatch(self) -> None: ...

        return _Endpoint

    def test_init(self, endpoint, app, route, asgi_scope, asgi_receive, asgi_send):
        asgi_scope = types.Scope(
            {
                **asgi_scope,
                "app": app,
                "root_app": app,
                "type": "http",
                "method": "GET",
                "path": "/",
                "path_params": {},
                "endpoint": endpoint,
                "route": route,
            }
        )
        e = endpoint(asgi_scope, asgi_receive, asgi_send)
        assert e.state == {
            "scope": asgi_scope,
            "receive": asgi_receive,
            "send": asgi_send,
            "exc": None,
            "app": app,
            "root_app": app,
            "path_params": {},
            "route": route,
        }

    def test_await(self, endpoint, app, route, asgi_scope, asgi_receive, asgi_send):
        e = endpoint(
            types.Scope(
                {
                    **asgi_scope,
                    "app": app,
                    "root_app": app,
                    "type": "http",
                    "method": "GET",
                    "path": "/",
                    "path_params": {},
                    "endpoint": endpoint,
                    "route": route,
                }
            ),
            asgi_receive,
            asgi_send,
        )

        with patch.object(e, "dispatch"):
            e.__await__()

            assert e.dispatch.call_args_list == [call()]



import re
import tempfile
import warnings
from contextlib import ExitStack
from pathlib import Path
from unittest.mock import AsyncMock

import pytest
from faker import Faker

import flama
from flama import Flama, types
from flama.client import Client
from flama.pagination import paginator
from flama.sqlalchemy import SQLAlchemyModule, metadata
from tests.utils import ExceptionContext, installed

try:
    import numpy as np
except Exception:
    warnings.warn("Numpy not installed")
    np = None

try:
    import sklearn.compose
    import sklearn.impute
    import sklearn.neural_network
    import sklearn.pipeline
except Exception:
    warnings.warn("SKLearn not installed")
    sklearn = None

try:
    import tensorflow as tf
except Exception:
    warnings.warn("Tensorflow not installed")
    tf = None

try:
    import torch
except Exception:
    warnings.warn("Torch not installed")
    torch = None

DATABASE_URL = "sqlite+aiosqlite://"


@pytest.fixture(scope="function")
def exception(request):
    if request.param is None:
        context = ExceptionContext(ExitStack())
    elif isinstance(request.param, Exception):
        context = ExceptionContext(
            pytest.raises(request.param.__class__, match=re.escape(str(request.param))), request.param
        )
    elif isinstance(request.param, (list, tuple)):
        exception, message = request.param
        context = ExceptionContext(pytest.raises(exception, match=re.escape(message)), exception)
    else:
        context = ExceptionContext(pytest.raises(request.param), request.param)
    return context


@pytest.fixture(scope="session")
def fake():
    return Faker()


@pytest.fixture(autouse=True)
def clear_metadata():
    metadata.clear()


@pytest.fixture(autouse=True)
def clear_pagination():
    paginator.schemas = {}


@pytest.fixture
def openapi_spec():
    return {"info": {"title": "Foo", "version": "1.0.0", "description": "Bar"}}


@pytest.fixture(
    scope="function",
    params=[
        pytest.param("pydantic", id="pydantic"),
        pytest.param("typesystem", id="typesystem"),
        pytest.param("marshmallow", id="marshmallow"),
    ],
)
def app(request, openapi_spec):
    return Flama(
        openapi=openapi_spec,
        schema="/schema/",
        docs="/docs/",
        modules={SQLAlchemyModule("sqlite+aiosqlite://")},
        schema_library=request.param,
    )


@pytest.fixture(scope="function")
async def client(app):
    async with Client(app=app) as client:
        assert client.app.status == types.AppStatus.READY
        yield client


@pytest.fixture(scope="function")
def asgi_scope():
    return {
        "app": Flama(),
        "type": "http",
        "method": "GET",
        "scheme": "https",
        "path": "/",
        "root_path": "",
        "query_string": b"",
        "headers": [],
    }


@pytest.fixture(scope="function")
def asgi_receive():
    return AsyncMock()


@pytest.fixture(scope="function")
def asgi_send():
    return AsyncMock()


class ModelFactory:
    def __init__(self):
        self._factories = {
            "sklearn": self._sklearn,
            "sklearn-pipeline": self._sklearn_pipeline,
            "tensorflow": self._tensorflow,
            "torch": self._torch,
        }

        self._models = {}
        self._models_cls = {}

    def _build(self, framework: str):
        if framework not in self._factories:
            raise ValueError(f"Wrong framework: '{framework}'.")

        if framework not in self._models:
            self._models[framework], self._models_cls[framework] = self._factories[framework]()

    def model(self, framework: str):
        self._build(framework)
        return self._models[framework]

    def model_cls(self, framework: str):
        self._build(framework)
        return self._models_cls[framework]

    def _sklearn(self):
        model = sklearn.neural_network.MLPClassifier(activation="tanh", max_iter=2000, hidden_layer_sizes=(10,))
        model.fit(
            np.array([[0, 0], [0, 1], [1, 0], [1, 1]]),
            np.array([0, 1, 1, 0]),
        )
        return model, sklearn.neural_network.MLPClassifier

    def _sklearn_pipeline(self):
        model = sklearn.neural_network.MLPClassifier(activation="tanh", max_iter=2000, hidden_layer_sizes=(10,))
        numerical_transformer = sklearn.pipeline.Pipeline(
            [
                ("imputer", sklearn.impute.SimpleImputer(strategy="constant", fill_value=0)),
            ]
        )
        preprocess = sklearn.compose.ColumnTransformer(
            [
                ("numerical", numerical_transformer[0, 1]),
            ]
        )
        pipeline = sklearn.pipeline.Pipeline(
            [
                ("preprocess", preprocess),
                ("model", model),
            ]
        )

        pipeline.fit(
            np.array([[0, np.nan], [np.nan, 1], [1, 0], [1, 1]]),  # NaN will be replaced with 0
            np.array([0, 1, 1, 0]),
        )

        return pipeline, sklearn.pipeline.Pipeline

    def _tensorflow(self):
        model = tf.keras.models.Sequential(
            [
                tf.keras.Input((2,)),
                tf.keras.layers.Dense(10, activation="tanh"),
                tf.keras.layers.Dense(1, activation="sigmoid"),
            ]
        )

        model.compile(optimizer="adam", loss="mse")
        model.fit(
            np.array([[0, 0], [0, 1], [1, 0], [1, 1]]),
            np.array([[0], [1], [1], [0]]),
            epochs=2000,
            verbose=0,
        )

        return model, tf.keras.models.Sequential

    def _torch(self):
        class Model(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.l1 = torch.nn.Linear(2, 10)
                self.l2 = torch.nn.Linear(10, 1)

            def forward(self, x):
                x = torch.tanh(self.l1(x))
                x = torch.sigmoid(self.l2(x))
                return x

            def _train(self, X, Y, loss, optimizer):
                for m in self.modules():
                    if isinstance(m, torch.nn.Linear):
                        m.weight.data.normal_(0, 1)

                steps = X.size(0)
                for i in range(2000):
                    for j in range(steps):
                        data_point = np.random.randint(steps)
                        x_var = torch.autograd.Variable(X[data_point], requires_grad=False)
                        y_var = torch.autograd.Variable(Y[data_point], requires_grad=False)

                        optimizer.zero_grad()
                        y_hat = model(x_var)
                        loss_result = loss.forward(y_hat, y_var)
                        loss_result.backward()
                        optimizer.step()

                return self

        X = torch.Tensor([[0, 0], [0, 1], [1, 0], [1, 1]])
        Y = torch.Tensor([0, 1, 1, 0]).view(-1, 1)
        model = Model()
        model._train(X, Y, loss=torch.nn.BCELoss(), optimizer=torch.optim.Adam(model.parameters()))

        return model, torch.jit.RecursiveScriptModule


model_factory = ModelFactory()


@pytest.fixture(scope="function")
def model(request):
    if not installed(request.param) or not installed("numpy"):
        pytest.skip(f"Lib '{request.param}' is not installed.")

    return model_factory.model(request.param)


@pytest.fixture(scope="function")
def serialized_model_class(request):
    if not installed(request.param) or not installed("numpy"):
        pytest.skip(f"Lib '{request.param}' is not installed.")

    return model_factory.model_cls(request.param)


@pytest.fixture(scope="function")
def model_path(request):
    if not installed(request.param) or not installed("numpy"):
        pytest.skip(f"Lib '{request.param}' is not installed.")

    with tempfile.NamedTemporaryFile(suffix=".flm") as f:
        flama.dump(model_factory.model(request.param), f.name)
        f.flush()

        yield Path(f.name)



import typing as t

from flama import exceptions
from flama.models.base import Model

try:
    import sklearn  # type: ignore
except Exception:  # pragma: no cover
    sklearn = None


class SKLearnModel(Model):
    def predict(self, x: list[list[t.Any]]) -> t.Any:
        if sklearn is None:  # noqa
            raise exceptions.FrameworkNotInstalled("scikit-learn")

        try:
            return self.model.predict(x).tolist()
        except ValueError as e:
            raise exceptions.HTTPException(status_code=400, detail=str(e))



from unittest.mock import AsyncMock, PropertyMock, call, patch

import pytest

from flama import http
from flama.debug.middleware import ExceptionMiddleware, ServerErrorMiddleware
from flama.middleware import Middleware, MiddlewareStack


class TestCaseMiddlewareStack:
    @pytest.fixture
    def middleware(self):
        class FooMiddleware:
            def __init__(self, *args, **kwargs): ...

            def __call__(self, *args, **kwargs):
                return None

        return Middleware(FooMiddleware)

    @pytest.fixture
    def app(self):
        return AsyncMock()

    @pytest.fixture
    def stack(self, app, middleware):
        return MiddlewareStack(app=app, middleware=[], debug=True)

    def test_init(self, app, middleware):
        stack = MiddlewareStack(app=app, middleware=[middleware], debug=True)

        assert stack.app == app
        assert stack.middleware == [middleware]
        assert stack.debug
        assert stack._exception_handlers == {}

    def test_stack(self, stack):
        assert stack._stack is None
        assert stack.stack
        assert isinstance(stack._stack, ServerErrorMiddleware)
        assert isinstance(stack._stack.app, ExceptionMiddleware)

    def test_add_exception_handler(self, stack):
        def handler(request: http.Request, exc: Exception) -> http.Response: ...

        assert stack._stack is None
        assert stack.stack
        assert stack._stack is not None

        stack.add_exception_handler(400, handler)
        stack.add_exception_handler(ValueError, handler)

        assert stack._stack is None
        assert stack._exception_handlers == {400: handler, ValueError: handler}

    def test_add_middleware(self, stack, middleware):
        assert stack._stack is None
        assert stack.stack
        assert stack._stack is not None

        stack.add_middleware(middleware)

        assert stack._stack is None
        assert stack.middleware == [middleware]

    async def test_call(self, stack, asgi_scope, asgi_receive, asgi_send):
        with patch.object(MiddlewareStack, "stack", new=PropertyMock(return_value=AsyncMock())):
            await stack(asgi_scope, asgi_receive, asgi_send)
            assert stack.stack.call_args_list == [call(asgi_scope, asgi_receive, asgi_send)]



import abc
import configparser
import json
import os
import typing as t

import yaml

from flama import compat, exceptions

__all__ = ["FileLoader", "ConfigFileLoader", "JSONFileLoader", "YAMLFileLoader", "TOMLFileLoader"]


class FileLoader(abc.ABC):
    """Common interface for loading a file."""

    @abc.abstractmethod
    def load(self, f: t.Union[str, os.PathLike]) -> dict[str, t.Any]:
        """Loads a file into a dict.

        :param f: File path.
        :return: Dict with the file contents.
        """
        ...


class ConfigFileLoader(FileLoader):
    """Loads an ini formatted file into a dict."""

    def load(self, f: t.Union[str, os.PathLike]) -> dict[str, t.Any]:
        """Loads a file into a dict.

        :param f: File path.
        :return: Dict with the file contents.
        """
        parser = configparser.ConfigParser()
        try:
            with open(f) as fs:
                parser.read_file(fs)
                return {section: dict(parser[section].items()) for section in parser.sections()}
        except configparser.MissingSectionHeaderError:
            with open(f) as fs:
                parser.read_string("[fake_section]\n" + fs.read())
                return dict(parser["fake_section"].items())


class JSONFileLoader(FileLoader):
    """Loads a json formatted file into a dict."""

    def load(self, f: t.Union[str, os.PathLike]) -> dict[str, t.Any]:
        """Loads a file into a dict.

        :param f: File path.
        :return: Dict with the file contents.
        """
        with open(f) as fs:
            return json.load(fs)


class YAMLFileLoader(FileLoader):
    """Loads a yaml formatted file into a dict."""

    def load(self, f: t.Union[str, os.PathLike]) -> dict[str, t.Any]:
        """Loads a file into a dict.

        :param f: File path.
        :return: Dict with the file contents.
        """
        with open(f) as fs:
            return yaml.safe_load(fs)


class TOMLFileLoader(FileLoader):
    """Loads a toml formatted file into a dict."""

    def load(self, f: t.Union[str, os.PathLike]) -> dict[str, t.Any]:
        """Loads a file into a dict.

        :param f: File path.
        :return: Dict with the file contents.
        """
        if compat.tomllib is None:  # PORT: Replace compat when stop supporting 3.10
            raise exceptions.DependencyNotInstalled(  # pragma: no cover
                dependency=exceptions.DependencyNotInstalled.Dependency.tomli,
                dependant=f"{self.__class__.__module__}.{self.__class__.__name__}",
                msg="for Python versions lower than 3.11",
            )

        with open(f, "rb") as fs:
            return compat.tomllib.load(fs)  # PORT: Replace compat when stop supporting 3.10



import dataclasses

import pytest

from flama.schemas.openapi import (
    Callback,
    Contact,
    Encoding,
    Example,
    ExternalDocs,
    Header,
    Info,
    License,
    Link,
    MediaType,
    OpenAPISpec,
    Operation,
    Parameter,
    Path,
    RequestBody,
    Response,
    Responses,
    Schema,
    Security,
    Server,
    ServerVariable,
)


class TestCaseOpenAPISpec:
    @pytest.fixture
    def spec(self):
        return OpenAPISpec(
            info=Info(
                title="Title",
                version="1.0.0",
                description="Description",
                termsOfService="TOS",
                contact=Contact(name="Contact name", url="https://contact.com", email="contact@contact.com"),
                license=License(name="License name", url="https://license.com"),
            )
        )

    @pytest.fixture
    def schema(self):
        return Schema(
            {
                "title": "Person",
                "type": "object",
                "properties": {
                    "firstName": {"type": "string", "description": "The person's first name."},
                    "lastName": {"type": "string", "description": "The person's last name."},
                    "age": {
                        "description": "Age in years which must be equal to or greater than zero.",
                        "type": "integer",
                        "minimum": 0,
                    },
                },
            }
        )

    @pytest.fixture
    def example(self, fake):
        return Example(summary=fake.sentence(), description=fake.text(), value={})

    @pytest.fixture
    def header(self, fake, schema, example):
        return Header(
            description=fake.text(),
            required=fake.pybool(),
            deprecated=fake.pybool(),
            allowEmptyValue=fake.pybool(),
            style=fake.sentence(),
            explode=fake.pybool(),
            allowReserved=fake.pybool(),
            schema=schema,
            example=example,
        )

    @pytest.fixture
    def server_variable(self, fake):
        return ServerVariable(enum=fake.pylist(value_types=[str]), default=fake.word(), description=fake.sentence())

    @pytest.fixture
    def server(self, fake, server_variable):
        return Server(url=fake.url(), variables={"var": server_variable}, description=fake.sentence())

    @pytest.fixture
    def link(self, fake, server):
        return Link(
            operationId=fake.word(),
            parameters=fake.pydict(value_types=[str]),
            requestBody={},
            description=fake.sentence(),
            server=server,
        )

    @pytest.fixture
    def encoding(self, fake, header):
        return Encoding(
            contentType="",
            headers={"header": header},
            style=fake.word(),
            explode=fake.pybool(),
            allowReserved=fake.pybool(),
        )

    @pytest.fixture
    def media_type(self, fake, schema, example, encoding):
        return MediaType(schema=schema, example=example, encoding={"enc": encoding})

    @pytest.fixture
    def response(self, header, media_type, link):
        return Response(
            description="Foo",
            headers={"header": header},
            content={"content": media_type},
            links={"link": link},
        )

    @pytest.fixture
    def parameter(self, fake, schema, example):
        return Parameter(
            name=fake.word(),
            in_="query",
            description=fake.sentence(),
            required=fake.pybool(),
            deprecated=fake.pybool(),
            allowEmptyValue=fake.pybool(),
            style=fake.word(),
            explode=fake.pybool(),
            allowReserved=fake.pybool(),
            schema=schema,
            example=example,
        )

    @pytest.fixture
    def request_body(self, fake, media_type):
        return RequestBody(content=media_type, description=fake.sentence(), required=fake.pybool())

    @pytest.fixture
    def security(self, fake):
        return Security({fake.word(): fake.pylist(value_types=[str])})

    @pytest.fixture
    def external_docs(self, fake):
        return ExternalDocs(url=fake.url(), description=fake.sentence())

    @pytest.fixture
    def operation(self, fake, response, external_docs, parameter, request_body, security, server):
        return Operation(
            responses=Responses({str(fake.random_number(digits=3)): response}),
            tags=fake.pylist(value_types=[str]),
            summary=fake.sentence(),
            description=fake.text(),
            externalDocs=external_docs,
            operationId=fake.word(),
            parameters=[parameter],
            requestBody=request_body,
            deprecated=fake.pybool(),
            security=[security],
            servers=[server],
        )

    @pytest.fixture
    def path(self, fake, operation, server, parameter):
        return Path(
            summary=fake.sentence(),
            description=fake.text(),
            get=operation,
            put=operation,
            post=operation,
            delete=operation,
            options=operation,
            head=operation,
            patch=operation,
            trace=operation,
            servers=[server],
            parameters=[parameter],
        )

    @pytest.fixture
    def callback(self, fake, path):
        return Callback({fake.word(): path})

    @pytest.mark.parametrize(
        ["spec", "result"],
        [
            pytest.param(
                {"info": {"title": "Title", "version": "1.0.0"}},
                {
                    "openapi": "3.1.0",
                    "info": {"title": "Title", "version": "1.0.0"},
                    "components": {
                        "callbacks": {},
                        "examples": {},
                        "headers": {},
                        "links": {},
                        "parameters": {},
                        "requestBodies": {},
                        "responses": {},
                        "schemas": {},
                        "securitySchemes": {},
                    },
                    "paths": {},
                },
                id="simple",
            ),
            pytest.param(
                {
                    "info": {
                        "title": "Example API",
                        "summary": "This is an example API specification",
                        "description": "A detailed description of the Example API, including usage and endpoints.",
                        "termsOfService": "https://example.com/terms",
                        "contact": {
                            "name": "API Support",
                            "url": "https://example.com/contact",
                            "email": "support@example.com",
                        },
                        "license": {
                            "name": "MIT",
                            "identifier": "MIT",
                            "url": "https://opensource.org/licenses/MIT",
                        },
                        "version": "1.0.0",
                    },
                    "servers": [
                        {
                            "url": "https://api.example.com/v1",
                            "description": "Production server",
                            "variables": {
                                "version": {
                                    "default": "v1",
                                    "enum": ["v1", "v2"],
                                    "description": "API version",
                                }
                            },
                        },
                        {
                            "url": "https://staging-api.example.com",
                            "description": "Staging server",
                        },
                    ],
                    "security": [
                        {
                            "OAuth2": ["read", "write"],
                            "ApiKeyAuth": [],
                        }
                    ],
                    "tags": [
                        {
                            "name": "users",
                            "description": "Operations related to users",
                            "externalDocs": {
                                "url": "https://docs.example.com/users",
                                "description": "User API documentation",
                            },
                        },
                        {
                            "name": "orders",
                            "description": "Operations related to orders",
                        },
                    ],
                    "externalDocs": {
                        "url": "https://docs.example.com",
                        "description": "Full API documentation",
                    },
                },
                {
                    "components": {
                        "callbacks": {},
                        "examples": {},
                        "headers": {},
                        "links": {},
                        "parameters": {},
                        "requestBodies": {},
                        "responses": {},
                        "schemas": {},
                        "securitySchemes": {},
                    },
                    "externalDocs": {
                        "description": "Full API documentation",
                        "url": "https://docs.example.com",
                    },
                    "info": {
                        "contact": {
                            "email": "support@example.com",
                            "name": "API Support",
                            "url": "https://example.com/contact",
                        },
                        "description": "A detailed description of the Example API, including usage and endpoints.",
                        "license": {
                            "identifier": "MIT",
                            "name": "MIT",
                            "url": "https://opensource.org/licenses/MIT",
                        },
                        "summary": "This is an example API specification",
                        "termsOfService": "https://example.com/terms",
                        "title": "Example API",
                        "version": "1.0.0",
                    },
                    "openapi": "3.1.0",
                    "paths": {},
                    "security": [
                        {
                            "ApiKeyAuth": [],
                            "OAuth2": [
                                "read",
                                "write",
                            ],
                        },
                    ],
                    "servers": [
                        {
                            "description": "Production server",
                            "url": "https://api.example.com/v1",
                            "variables": {
                                "version": {
                                    "default": "v1",
                                    "description": "API version",
                                    "enum": [
                                        "v1",
                                        "v2",
                                    ],
                                },
                            },
                        },
                        {
                            "description": "Staging server",
                            "url": "https://staging-api.example.com",
                        },
                    ],
                    "tags": [
                        {
                            "description": "Operations related to users",
                            "externalDocs": {
                                "description": "User API documentation",
                                "url": "https://docs.example.com/users",
                            },
                            "name": "users",
                        },
                        {
                            "description": "Operations related to orders",
                            "name": "orders",
                        },
                    ],
                },
                id="full",
            ),
        ],
    )
    def test_from_spec_to_dict(self, spec, result):
        assert OpenAPISpec.from_spec(spec).to_dict() == result
        assert OpenAPISpec(info=Info(title="Title", version="1.0.0"))
        assert OpenAPISpec(
            info=Info(
                title="Title",
                version="1.0.0",
                description="Description",
                termsOfService="TOS",
                contact=Contact(name="Contact name", url="Contact url", email="Contact email"),
                license=License(name="License name", url="License url"),
            ),
        )

    def test_add_path(self, spec, path):
        path_name = "foo"
        expected_result = spec.to_dict()
        expected_result["paths"] = {"foo": spec.to_dict(dataclasses.asdict(path))}

        spec.add_path(path_name, path)

        assert spec.to_dict() == expected_result

    def test_add_schema(self, spec, schema):
        schema_name = "foo"
        expected_result = spec.to_dict()
        expected_result["components"]["schemas"] = {schema_name: schema}

        spec.add_schema(schema_name, schema)

        assert spec.to_dict() == expected_result

    def test_add_response(self, spec, response):
        response_name = "foo"
        expected_result = spec.to_dict()
        expected_result["components"]["responses"] = {response_name: spec.to_dict(dataclasses.asdict(response))}

        spec.add_response(response_name, response)

        assert spec.to_dict() == expected_result

    def test_add_parameter(self, spec, parameter):
        parameter_name = "foo"
        expected_result = spec.to_dict()
        expected_result["components"]["parameters"] = {parameter_name: spec.to_dict(dataclasses.asdict(parameter))}

        spec.add_parameter(parameter_name, parameter)

        assert spec.to_dict() == expected_result

    def test_add_example(self, spec, example):
        example_name = "foo"
        expected_result = spec.to_dict()
        expected_result["components"]["examples"] = {example_name: spec.to_dict(dataclasses.asdict(example))}

        spec.add_example(example_name, example)

        assert spec.to_dict() == expected_result

    def test_add_request_body(self, spec, request_body):
        request_body_name = "foo"
        expected_result = spec.to_dict()
        expected_result["components"]["requestBodies"] = {
            request_body_name: spec.to_dict(dataclasses.asdict(request_body))
        }

        spec.add_request_body(request_body_name, request_body)

        assert spec.to_dict() == expected_result

    def test_add_header(self, spec, header):
        header_name = "foo"
        expected_result = spec.to_dict()
        expected_result["components"]["headers"] = {header_name: spec.to_dict(dataclasses.asdict(header))}

        spec.add_header(header_name, header)

        assert spec.to_dict() == expected_result

    def test_add_security(self, spec, security):
        security_name = "foo"
        expected_result = spec.to_dict()
        expected_result["components"]["securitySchemes"] = {security_name: spec.to_dict(security)}

        spec.add_security(security_name, security)

        assert spec.to_dict() == expected_result

    def test_add_link(self, spec, link):
        link_name = "foo"
        expected_result = spec.to_dict()
        expected_result["components"]["links"] = {link_name: spec.to_dict(dataclasses.asdict(link))}

        spec.add_link(link_name, link)

        assert spec.to_dict() == expected_result

    def test_add_callback(self, spec, callback):
        callback_name = "foo"
        expected_result = spec.to_dict()
        expected_result["components"]["callbacks"] = {callback_name: spec.to_dict(callback)}

        spec.add_callback(callback_name, callback)

        assert spec.to_dict() == expected_result



import typing as t

from flama import types

__all__ = ["BaseEndpoint"]


class BaseEndpoint(types.EndpointProtocol):
    def __init__(self, scope: "types.Scope", receive: "types.Receive", send: "types.Send") -> None:
        """An endpoint.

        :param scope: ASGI scope.
        :param receive: ASGI receive function.
        :param send: ASGI send function.
        """
        app = scope["app"]
        scope["path"] = scope.get("root_path", "").rstrip("/") + scope["path"]
        scope["root_path"] = ""
        route, route_scope = app.router.resolve_route(scope)
        self.state = {
            "scope": route_scope,
            "receive": receive,
            "send": send,
            "exc": None,
            "app": app,
            "root_app": scope["root_app"],
            "path_params": route_scope.get("path_params", {}),
            "route": route,
        }

    def __await__(self) -> t.Generator:
        return self.dispatch().__await__()



import asyncio

import flama
from flama import Flama, background
from flama.http import JSONResponse

app = Flama()


async def sleep_task(value: int):
    await asyncio.sleep(value)


@app.route("/")
async def test():
    task = background.BackgroundThreadTask(sleep_task, 10)
    return JSONResponse("hello", background=task)


if __name__ == "__main__":
    flama.run(flama_app=app, server_host="0.0.0.0", server_port=8080)



class ResourceAttributeError(AttributeError):
    ATTRIBUTE_NOT_FOUND = "needs to define attribute '{attribute}'"
    # RESTResource
    SCHEMA_NOT_FOUND = "needs to define attribute 'schema' or the pair 'input_schema' and 'output_schema'"
    RESOURCE_NAME_INVALID = "invalid resource name '{resource_name}'"
    PK_NOT_FOUND = "model must define a single-column primary key"
    PK_WRONG_TYPE = "model primary key wrong type"
    MODEL_INVALID = "model must be a valid SQLAlchemy Table instance or a Model instance"
    # ModelResource
    MODEL_NOT_FOUND = "needs to define attribute 'model_path' or 'component'"

    def __init__(self, msg: str, name: str):
        super().__init__(f"{name} {msg}")



import abc
import asyncio
import logging
import typing as t

from flama.ddd.repositories import BaseRepository
from flama.exceptions import ApplicationError

if t.TYPE_CHECKING:
    from flama import Flama

logger = logging.getLogger(__name__)

__all__ = ["AbstractWorker", "BaseWorker"]


Repositories = t.NewType("Repositories", dict[str, type[BaseRepository]])


class AbstractWorker(abc.ABC):
    """Abstract class for workers.

    It will be used to define the workers for the application. A worker must provide a mechanism to isolate a single
    unit of work that will be used to interact with the repositories and entities of the application.
    """

    def __init__(self, app: t.Optional["Flama"] = None):
        """Initialize the worker.

        It will receive the application instance as a parameter.

        :param app: Application instance.
        """
        self._app = app
        self._lock = asyncio.Lock()

    @property
    def app(self) -> "Flama":
        """Application instance.

        :return: Application instance.
        """
        if not self._app:
            raise ApplicationError("Worker not initialized")

        return self._app

    @app.setter
    def app(self, app: "Flama") -> None:
        """Set the application instance.

        :param app: Application instance.
        """
        self._app = app

    @app.deleter
    def app(self) -> None:
        """Delete the application instance."""
        self._app = None

    @abc.abstractmethod
    async def begin(self) -> None:
        """Start a unit of work."""
        ...

    @abc.abstractmethod
    async def end(self, *, rollback: bool = False) -> None:
        """End a unit of work.

        :param rollback: If the unit of work should be rolled back.
        """
        ...

    async def __aenter__(self) -> "AbstractWorker":
        """Start a unit of work."""
        await self._lock.acquire()
        logger.debug("Start unit of work")
        await self.begin()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """End a unit of work."""
        await self.end(rollback=exc_type is not None)
        logger.debug("End unit of work")
        self._lock.release()

    @abc.abstractmethod
    async def commit(self) -> None:
        """Commit the unit of work."""
        ...

    @abc.abstractmethod
    async def rollback(self) -> None:
        """Rollback the unit of work."""
        ...


class WorkerType(abc.ABCMeta):
    """Metaclass for workers.

    It will gather all the repositories defined in the class as class attributes as a single dictionary under the name
    `_repositories` and remove them from the class annotations.
    """

    def __new__(mcs, name: str, bases: tuple[type], namespace: dict[str, t.Any]):
        if not mcs._is_base(namespace) and "__annotations__" in namespace:
            namespace["_repositories"] = Repositories(
                {k: v for k, v in namespace["__annotations__"].items() if mcs._is_repository(v)}
            )

            namespace["__annotations__"] = {
                k: v for k, v in namespace["__annotations__"].items() if k not in namespace["_repositories"]
            }

        return super().__new__(mcs, name, bases, namespace)

    @staticmethod
    def _is_base(namespace: dict[str, t.Any]) -> bool:
        return namespace.get("__module__") == "flama.ddd.workers" and namespace.get("__qualname__") == "BaseWorker"

    @staticmethod
    def _is_repository(obj: t.Any) -> bool:
        try:
            return issubclass(obj, BaseRepository)
        except TypeError:
            return False


class BaseWorker(AbstractWorker, metaclass=WorkerType):
    """Base class for workers.

    It will be used to define the workers for the application. A worker consists of a set of repositories that will be
    used to interact with entities and a mechanism for isolate a single unit of work.

    It will gather all the repositories defined in the class as class attributes as a single dictionary under the name
    `_repositories` and remove them from the class annotations.
    """

    _repositories: t.ClassVar[dict[str, type[BaseRepository]]]

    @abc.abstractmethod
    async def set_up(self) -> None:
        """First step in starting a unit of work."""
        ...

    @abc.abstractmethod
    async def tear_down(self, *, rollback: bool = False) -> None:
        """Last step in ending a unit of work.

        :param rollback: If the unit of work should be rolled back.
        """
        ...

    @abc.abstractmethod
    async def repository_params(self) -> tuple[list[t.Any], dict[str, t.Any]]:
        """Get the parameters for initialising the repositories.

        :return: Parameters for initialising the repositories.
        """
        ...

    async def begin(self) -> None:
        """Start a unit of work."""
        await self.set_up()

        args, kwargs = await self.repository_params()

        for repository, repository_class in self._repositories.items():
            setattr(self, repository, repository_class(*args, **kwargs))

    async def end(self, *, rollback: bool = False) -> None:
        """End a unit of work.

        :param rollback: If the unit of work should be rolled back.
        """
        await self.tear_down(rollback=rollback)

        for repository in self._repositories.keys():
            delattr(self, repository)
